from _collections_abc import dict_keys, dict_values
from pyg_base._as_list import as_list, as_tuple
from pyg_base._as_float import as_float
from pyg_base._dict import Dict
from pyg_base._zip import zipper, lens
from pyg_base._types import is_str, is_strs, is_arr, is_df, is_dicts, is_int, is_ints, is_tuple, is_bools, is_nan
from pyg_base._tree import is_tree, tree_to_table
from pyg_base._inspect import getargs
from pyg_base._decorators import kwargs_support, try_none, try_back
from pyg_base._dates import ndt
from pyg_base._sort import sort, cmp
from pyg_base._encode import _encode, _obj
from pyg_base._file import read_csv
from functools import reduce
import pandas as pd
import re

__all__ = ['dict_concat', 'dictable', 'is_dictable']


def nan2none(v):
    return None if is_nan(v) or (is_str(v) and (len(v.strip()) == 0 or v[0] == '#' or v == 'n/a')) else v


def dict_concat(*dicts):
    """
    A method of turning a list of dicts into one happy dict sharing its keys

    :Parameters:
    ----------------
    *dicts : list of dicts
        a list of dicts

    :Returns:
    -------
    dict
        a dict of a list of values.
    
    :Example:
    --------------
    >>> dicts = [dict(a=1,b=2), dict(a=3,b=4,c=5)]
    >>> assert dict_concat(dicts) == dict(a = [1,3], b = [2,4], c = [None,5])
    >>> dicts = [dict(a=1,b=2)]
    >>> assert dict_concat(dicts) == dict(a = [1], b = [2])
    >>> assert dict_concat([]) == dict()
    """
    
    dicts = as_list(dicts)
    if len(dicts) == 0:
        return {}
    elif len(dicts) == 1:
        return {key: [value] for key, value in dicts[0].items()}  # a shortcut, we dont need to run a full merge
    possible_keys = list(set([tuple(sorted(d.keys())) for d in dicts]))
    if len(possible_keys) == 1:
        pairs = [sorted(d.items()) for d in dicts]
        keys = possible_keys[0]
        values = zip(*[[value for _, value in row] for row in pairs])
        res = dict(zip(keys, map(list, values)))
        return res
    else:
        keys = reduce(lambda res, keys: res | set(keys), possible_keys, set())
        return {key: [d.get(key) for d in dicts] for key in keys}
    

def _text_box(value, max_rows = 5, max_chars = 50):
    """
    a = dictable(a = range(10)) * dict(b = range(10))
    b = (dictable(a = range(5)) * dict(b = range(5))) * dict(c = range(10))
    self = c = dictable(key = ['a', 'b'] * 5, rs = [a,b]*5)
    """
    v = str(value)
    res = v.split('\n')
    if max_rows:
        res = res[:max_rows]
    if max_chars:
        res = [row[:max_chars] for row in res]
    return res


def _data_columns_as_dict(data, columns = None):
    """
    >>> assert _data_columns_as_dict(data = [], columns = []) == dict()
    >>> assert _data_columns_as_dict([], 'a') == dict(a = [])
    >>> assert _data_columns_as_dict(data = [], columns = ['a','b']) == dict(a = [], b = [])
    
    :Parameters:
    ----------------
    data : TYPE
        DESCRIPTION.
    columns : TYPE, optional
        DESCRIPTION. The default is None.

    :Returns:
    -------
    TYPE
        DESCRIPTION.

    """
    if isinstance(data, zip):
        data = list(data)
    if data is None or (isinstance(data, list) and data == []):
        return {}
    elif isinstance(data, dict):
        return dict_concat(tree_to_table(data, columns)) if is_tree(columns) else dict(data)
    if columns is not None:
        if is_str(columns):
            if (is_str(data) and '.xls' in data) or isinstance(data, pd.io.excel.ExcelFile):
                return dict(dictable.read_excel(data, columns))
            else:
                return {columns : data}
        else:
            return dict(zipper(columns, zipper(*data)))
    else:
        if is_str(data):
            if data.endswith('.csv'):
                data = read_csv(data)
            elif 'xls' in data:
                data = dict(dictable.read_excel(data))
            else:
                return dict(data = data)
        if is_df(data):
            if data.index.name is not None:
                data = data.reset_index()
            return data.to_dict('list')
        tp = str(type(data)).lower()        
        if hasattr(data, 'next') and 'cursor' in tp:
            data = list(data)
        elif hasattr(data, 'find') and 'collection' in tp:
            data = list(data.find({}))
        if isinstance(data, list):
            if len(data) == 0:
                return {}
            elif min([is_tuple(row) and len(row) == 2 for row in data]):
                return dict(data)
            elif is_dicts(data):
                return dict_concat(data)
            elif min([isinstance(i, list) for i in data]):
                return dict(zipper(data[0], zipper(*data[1:])))
            else:
                return dict(data = data)
        elif isinstance(data, dict):
            return dict(data)
        else:
            return dict(data = data)


def _value(value):
    if value is None:
        return [None]
    elif isinstance(value, (dict_values, dict_keys, range)):
        return list(value)
    else:
        return as_list(value)

class dictable(Dict):
    """
    :What is dictable?:
    -------------------
    dictable is a table, a collection of iterable records. It is also a dict with each key being a column. 
    Why not use a pandas.DataFrame? pd.DataFrame leads a dual life: 
        - by day an index-based optimized numpy array supporting e.g. timeseries analytics etc.
        - by night, a table with keys supporting filtering, aggregating, pivoting on keys as well as inner/outer joining on keys.

    
    dictable only tries to do the latter. dictable should be thought of as a 'container for complicated objects' rather than just an array of primitive floats.
    In general, each cell may contain timeseries, yield_curves, machine-learning experiments etc.
    The interface is very succinct and allows the user to concentrate on logic of the calculations rather than boilerplate.
    
    dictable supports quite a flexible construction:

    :Example: construction using records
    ------------------------------------
    >>> from pyg import *; import pandas as pd
    >>> d = dictable([dict(name = 'alan', surname = 'atkins', age = 39, country = 'UK'), 
    >>>               dict(name = 'barbara', surname = 'brown', age = 29, country = 'UK')])
    
    :Example: construction using columns and constants
    ---------------------------------------------------
    >>> d = dictable(name = ['alan', 'barbara'], surname = ['atkins', 'brown'], age = [39, 29], country = 'UK')
    
    :Example: construction using pandas.DataFrame
    ---------------------------------------------
    >>> original = dictable(name = ['alan', 'barbara'], surname = ['atkins', 'brown'], age = [39, 29], country = 'UK')
    >>> df_from_dictable = pd.DataFrame(original)
    >>> dictable_from_df = dictable(df_from_dictable)
    >>> assert original == dictable_from_df

    :Example: construction rows and columns
    ---------------------------------------
    >>> d = dictable([['alan', 'atkins', 39, 'UK'], ['barbara', 'brown', 29, 'UK']], columns = ['name', 'surname', 'age', 'country'])


    :Access: column access
    ----------------------
    >>> assert d.keys() ==  ['name', 'surname', 'age', 'country']
    >>> assert d.name == ['alan', 'barbara']
    >>> assert d['name'] == ['alan', 'barbara']
    >>> assert d['name', 'surname'] == [('alan', 'atkins'), ('barbara', 'brown')]
    >>> assert d[lambda name, surname: '%s %s'%(name, surname)] == ['alan atkins', 'barbara brown']


    :Access: row access & iteration
    -------------------------------
    >>> assert d[0] == {'name': 'alan', 'surname': 'atkins', 'age': 39, 'country': 'UK'}
    >>> assert [row for row in d] == [{'name': 'alan', 'surname': 'atkins', 'age': 39, 'country': 'UK'},
    >>>                               {'name': 'barbara', 'surname': 'brown', 'age': 29, 'country': 'UK'}]

    Note that members access is commutative: 

    >>> assert d.name[0] == d[0].name == 'alan'
    >>> d[lambda name, surname: name + surname][0] == d[0][lambda name, surname: name + surname]
    >>> assert sum([row for row in d], dictable()) == d

    :Example: adding records
    ------------------------
    >>> d = dictable(name = ['alan', 'barbara'], surname = ['atkins', 'brown'], age = [39, 29], country = 'UK')
    >>> d = d + {'name': 'charlie', 'surname': 'chocolate', 'age': 49} # can add a record directly
    >>> assert d[-1] == {'name': 'charlie', 'surname': 'chocolate', 'age': 49, 'country': None}
    >>> d += dictable(name = ['dana', 'ender'], surname = ['deutch', 'esterhase'], age = [10, 20], country = ['Germany', 'Hungary'])
    >>> assert d.name == ['alan', 'barbara', 'charlie', 'dana', 'ender']
    >>> assert len(dictable.concat([d,d])) == len(d) * 2

    :Example: adding columns
    ------------------------
    >>> d = dictable(name = ['alan', 'barbara'], surname = ['atkins', 'brown'], age = [39, 29], country = 'UK')
    
    >>> ### all of the below are ways of adding columns ####
    >>> d.gender == ['m', 'f']
    >>> d = d(gender = ['m', 'f'])
    >>> d['gender'] == ['m', 'f']
    >>> d2 = dictable(gender = ['m', 'f'], profession = ['astronaut', 'barber'])
    >>> d = d(**d2)

    :Example: adding derived columns
    --------------------------------
    >>> d = dictable(name = ['alan', 'barbara'], surname = ['atkins', 'brown'], age = [39, 29], country = 'UK')
    >>> d = d(full_name = lambda name, surname: proper('%s %s'%(name, surname))) 
    >>> d['full_name'] = d[lambda name, surname: proper('%s %s'%(name, surname))]
    >>> assert d.full_name == ['Alan Atkins', 'Barbara Brown']

    :Example: dropping columns
    ---------------------------
    >>> d = dictable(name = ['alan', 'barbara'], surname = ['atkins', 'brown'], age = [39, 29], country = 'UK')
    >>> del d.country # in place
    >>> del d['age'] # in place
    >>> assert (d - 'name')[0] ==  {'surname': 'atkins'} and d[0] == {'name': 'alan', 'surname': 'atkins'}

    :Example: row selection, inc/exc
    --------------------------------------
    >>> d = dictable(name = ['alan', 'barbara'], surname = ['atkins', 'brown'], age = [39, 29], country = 'UK')
    >>> assert len(d.exc(name = 'alan')) == 1
    >>> assert len(d.exc(lambda age: age<30)) == 1 # can filter on *functions* of members, not just members.
    >>> assert d.inc(name = 'alan').surname == ['atkins']
    >>> assert d.inc(lambda age: age<30).name == ['barbara']
    >>> assert d.exc(lambda age: age<30).name == ['alan']

    dictable supports:
        - sort 
        - group-by/ungroup
        - list-by/ unlist
        - pivot/unpivot
        - inner join, outer join and xor

    Full details are below.
    """
    def __init__(self, data = None, columns = None, **kwargs):
        kwargs = {key :_value(value) for key, value in kwargs.items()}
        data_kwargs = {key: _value(value) for key, value in _data_columns_as_dict(data, columns).items()}
        kwargs.update(data_kwargs)
        if is_strs(columns) and (len(data_kwargs) == 0 or not is_str(columns)):
            kwargs = {key : kwargs.get(key, [None]) for key in columns} if len(kwargs)>0 else {key : [] for key in columns}
        n = lens(*kwargs.values())
        kwargs = {str(key) if is_int(key) else key : value * n if len(value)==1 else value for key, value in kwargs.items()}
        super(dictable, self).__init__(kwargs)
        
    def __len__(self):
        return lens(*self.values())
    
    @property
    def shape(self):
        return (len(self), len(self.keys()))
    
    @property
    def columns(self):
        return self.keys()

    def get(self, key, default = None):
        if key in self:
            return self[key]
        else:
            return [default] * len(self)
        
    def __iter__(self):
        for row in zip(*self.values()):
            yield Dict(zip(self.keys(), row))
            
    def update(self, other):
        for k, v in other.items():
            self[k] = v
    
    def __setitem__(self, key, value):
        n = len(self)
        value = _value(value)
        if len(value) == n or len(self.keys()) == 0:
            pass
        elif len(value) == 1:
            value = value * n 
        else:
            raise ValueError('cannot set item of length %s in table of length %s'%(len(value), n))
        super(dictable, self).__setitem__(str(key) if is_int(key) else key, value)
    
    def __getitem__(self, item):
        if is_arr(item) and len(item.shape) == 1:
            item = list(item)
        if isinstance(item, slice):
            return type(self)({key : value[item] for key, value in self.items()})
        if isinstance(item, (dict_keys, dict_values, range)):
            item = list(item)
        if isinstance(item , list):
            if len(item) == 0:
                return dictable(data = [], columns = self.keys())
            elif is_strs(item):
                return type(self)(super(dictable, self).__getitem__(item))
            elif is_bools(item):
                res = type(self)([row for row, tf in zipper(self, item) if tf])
                return res if len(res) else type(self)([], self.keys())
            elif is_ints(item):
                values = list(zip(*self.values()))
                return dictable(data = [values[i] for i in item], columns = self.keys())
            else:
                raise ValueError('We dont know how to understand this item %s'%item)
        elif is_int(item):
            return Dict({key : value[item] for key, value in self.items()})
        elif item in self.keys():
            return super(dictable, self).__getitem__(item)
        elif is_tuple(item):
            return list(zip(*[self[i] for i in item]))
        elif callable(item):
            return self.apply(item)
        else:
            raise KeyError('item %s not found'%item)

    def __getattr__(self, attr):
        if attr in self:
            return self[attr]
        elif attr.startswith('_'):
            return super(dictable, self).__getattr__(attr)
        elif attr.startswith('find_'):
            key = attr[5:]
            if key not in self:
                raise KeyError('cannot find %s in dictable'%key)
            def f(*args, **kwargs):
                items = self.inc(*args, **kwargs)
                if len(items) == 0:
                    raise ValueError('no %s found'%key)
                item = items[key]
                if len(item)>1:
                    item = list(set(item))
                    if len(item)>1:
                        raise ValueError('multiple %s found %s'%(key, item))
                return item[0]
            f.__name__ = 'find_%s'%key
            return f
        else:
            raise AttributeError('%s not found'%attr)
 
    def __setattr__(self, attr, value):
        if attr.startswith('_'):
            super(dictable, self).__setattr__(attr, value)
        else:
            self.__setitem__(attr, value)

    def __delattr__(self, attr):
        if attr.startswith('_'):
            super(dictable, self).__delattr__(attr)
        else:
            super(dictable, self).__delitem__(attr)
    
    def _encode(self):
        """
        We override the encoding method for dictable by enforcing columns to be specified as part of the dict.
        This ensures that when constructed from mongodb, old columns that may still be part of the cell are not included in the construction

        """
        res = {k : _encode(v) for k, v in self.items()}
        if _obj not in res:
            res[_obj] = _encode(type(self))
        res['columns'] = self.columns
        return res

    def inc(self, *functions, **filters):
        """
        performs a filter on what rows to include

        :Parameters:
        ----------------
        *functions : callables or a dict
            filters based on functions of each row
        **filters : value or list of values
            filters per each column

        :Returns:
        -------
        dictable
            table with rows that satisfy all conditions.
            
        
        :Example: filtering on keys
        -------
        >>> from pyg import *; import numpy as np
        >>> d = dictable(x = [1,2,3,np.nan], y = [0,4,3,5])
        >>> assert d.inc(x = np.nan) == dictable(x = np.nan, y = 5)            
        >>> assert d.inc(x = 1) == dictable(x = 1, y = 0)            
        >>> assert d.inc(x = [1,2]) == dictable(x = [1,2], y = [0,4]) 

        :Example: filtering on regex
        -------
        >>> import re
        >>> d = dictable(text = ['once', 'upon', 'a', 'time', 'in', 'the', 'west', 1, 2, 3])
        >>> assert d.inc(text = re.compile('o')) == dictable(text = ['once', 'upon'])
        >>> assert d.exc(text = re.compile('e')) == dictable(text = ['upon', 'a', 'in', 1, 2, 3])
        
        
        :Example: filtering on callables
        --------------
        >>> from pyg import *; import numpy as np
        >>> d = dictable(x = [1,2,3,np.nan], y = [0,4,3,5])
        >>> assert d.inc(lambda x,y: x>y) == dictable(x = 1, y = 0)

        """
        res = self.copy()
        if len(functions) + len(filters) == 0:
            return res
        functions = as_list(functions)
        for function in functions:
            if type(function) == dict:
                filters.update(function)
            else:
                f = kwargs_support(function)
                res = type(self)([row for row in res if f(**row)])
        for key, value in filters.items():
            if value is None:
                res = res[[r is None for r in res[key]]]
            elif is_nan(value):
                res = res[[is_nan(r) for r in res[key]]]
            elif isinstance(value, re.Pattern):
                res = res[[is_str(r) and value.search(r) is not None for r in res[key]]]                
            else:
                value = as_list(value)
                res = res[[r in value for r in res[key]]]
        if len(res) == 0:
            return type(self)([], self.keys())
        return res                

    def exc(self, *functions, **filters):
        """
        performs a filter on what rows to exclude

        :Parameters:
        ----------------
        *functions : callables or a dict
            filters based on functions of each row
        **filters : value or list of values
            filters per each column

        :Returns:
        -------
        dictable
            table with rows that satisfy all conditions excluded.
            
        
        :Example: filtering on keys
        -------
        >>> from pyg import *; import numpy as np
        >>> d = dictable(x = [1,2,3,np.nan], y = [0,4,3,5])
        >>> assert d.exc(x = np.nan) == dictable(x = [1,2,3], y = [0,4,3])         
        >>> assert d.exc(x = 1) == dictable(x = [2,3,np.nan], y = [4,3,5])
        >>> assert d.exc(x = [1,2]) == dictable(x = [1,2], y = [0,4]) 

        :Example: filtering on callables
        --------------
        >>> from pyg import *; import numpy as np
        >>> d = dictable(x = [1,2,3,np.nan], y = [0,4,3,5])
        >>> assert d.exc(lambda x,y: x>y) == dictable(x = 1, y = 0)

        """
        res = self.copy()
        if len(functions) + len(filters) == 0:
            return res
        functions = as_list(functions)
        for function in functions:
            if type(function) == dict:
                filters.update(function)
            else:
                f = kwargs_support(function)
                res = type(self)([row for row in res if not f(**row)])
        for key, value in filters.items():
            if value is None:
                res = res[[r is not None for r in res[key]]]
            elif is_nan(value):
                res = res[[not is_nan(r) for r in res[key]]]
            elif isinstance(value, re.Pattern):
                res = res[[(not is_str(r)) or value.search(r) is None for r in res[key]]]                
            else:
                value = as_list(value)
                res = res[[r not in value for r in res[key]]]
        if len(res) == 0:
            return type(self)([], self.keys())
        return res                


    def if_none(self, none = None, **kwargs):
        """
        runs a column calculation if a column is not there, or if the existing value is considered missing.
        
        Example:
        --------
        >>> rs = dictable(a = [1,'hello'])
        >>> rs = rs(b = try_none(lambda a: a[0]))  ## fails for numerical value
        >>> rs

        >>> dictable[2 x 2]
        >>> a    |b   
        >>> 1    |None
        >>> hello|h   
        
        what shall we do with integer values for which b is None?
        
        >>> rs = rs.if_none(b = lambda a: a+3)
        >>> rs
        >>> dictable[2 x 2]
        >>> a    |b   
        >>> 1    |4
        >>> hello|h   


        Example:
        -------- 
        >>> rs = dictable(a = range(3))
        >>> rs = rs(b = try_nan(lambda a: 10/a))
        >>> rs = rs.if_none(np.nan, b = 'sorry no can do')

        >>> dictable[3 x 2]
        >>> a|b              
        >>> 0|sorry no can do
        >>> 1|10.0           
        >>> 2|5.0        
        
        """
        if none is None:
            check = lambda value: value is None
        elif not callable(none):
            check = lambda value: value in as_list(none)
        else:
            check = none
        res = self
        for key, value in kwargs.items():
            if key not in res.keys():
                res[key] = res[value]
            else:
                if callable(value):
                    res[key] = [row[key] if not check(row[key]) else row[value] for row in res]
                else:
                    res[key] = [row[key] if not check(row[key]) else value for row in res]                    
        return res

    
    def apply(self, function):
        f = kwargs_support(function)
        return [f(**row) for row in self]

    def do(self, function, *keys):
        """
        applies a function(s) on multiple keys at the same time

        :Parameters:
        ----------------
        function : callable or list of callables
            function to be applied per column
        *keys : string/list of strings
            list of columns to be applied. If missing, applied to all columns

        :Returns:
        -------
        res : dictable

        :Example:
        --------------
        >>> from pyg import *
        >>> d = dictable(name = ['adam', 'barbara', 'chris'], surname = ['atkins', 'brown', 'cohen'])
        >>> assert d.do(proper) == dictable(name = ['Adam', 'Barbara', 'Chris'], surname = ['Atkins', 'Brown', 'Cohen'])

        :Example: using another column in the calculation
        -------
        >>> from pyg import *
        >>> d = dictable(a = [1,2,3,4], b = [5,6,9,8], denominator = [10,20,30,40])
        >>> d = d.do(lambda value, denominator: value/denominator, 'a', 'b')
        >>> assert d == dictable(a = 0.1, b = [0.5,0.3,0.3,0.2], denominator = [10,20,30,40])
        
        """
        res = self.copy()
        keys = as_list(keys)
        if len(keys)  == 0:
            keys = self.keys()
        for key in keys:    
            for f in as_list(function):
                args = as_list(try_none(getargs)(f))
                res[key] = [f(row[key], **{k : v for k, v in row.items() if k in args[1:]}) for row in res]
        return res

    @classmethod
    def concat(cls, *others):
        """
        adds together multiple dictables. equivalent to sum(others, self) but a little faster

        :Parameters:
        ----------------
        *others : dictables
            records to be added to current table

        :Returns:
        -------
        merged : dictable
            sum of all records
            
        :Example:
        -------
        >>> from pyg import *
        >>> d1 = dictable(a = [1,2,3])
        >>> d2 = dictable(a = [4,5,6])
        >>> d3 = dictable(a = [7,8,9])
        
        >>> assert dictable.concat(d1,d2,d3) == dictable(a = range(1,10))
        >>> assert dictable.concat([d1,d2,d3]) == dictable(a = range(1,10))
        """
        others = as_list(others)
        others = [cls(other) if not isinstance(other, cls) else other for other in others]
        if len(others) == 0:
            return cls()
        elif len(others) == 1:
            return others[0]
        concated = dict_concat(others)
        merged = cls({key : sum(value, []) for key, value in concated.items()}) 
        return merged

    def __repr__(self):
        return 'dictable[%s x %s]'%self.shape + '\n%s'%self.__str__(3, 150)
    
    
    def __str__(self, max_rows = None, max_width = 150):
        return self.to_string(cat = max_rows, mid = '...%i rows...'%len(self), max_width=max_width)
    
    def to_string(self, colsep = '|', rowsep = '', cat = None, max_rows = 5, max_width = None, mid = '...', header = None, footer = None):
        """
        convrts dictable to string
        :Example:
        ---------
        >>> colsep = '|'; rowsep = ''; cat = None; max_rows = 5; max_width = None; mid = '...'; header = None; footer = None
        
        """
        if cat and len(self) > 2 * cat:
            box = {key: [_text_box(key)] + 
                        [_text_box('dictable[%s x %s]\n%s'%(len(v), len(v.keys()), v.keys()) if isinstance(v, dictable) else v, max_rows) for v in value[:cat]+value[-cat:]]
                    for key, value in self.items()}
        else:
            box = {key: [_text_box(key)] + 
                        [_text_box('dictable[%s x %s]\n%s'%(len(v), len(v.keys()), v.keys()) if isinstance(v, dictable) else v, max_rows) for v in value]
                    for key, value in self.items()}
    
        chars = {key : max([max([len(r) for r in v]) for v in value]) for key, value in box.items()}
        padded = {key : [[r + ' ' * (chars[key]-len(r)) for r in v] for v in value] for key, value in box.items()}
        def text(padded, keys = None):
            padded_ = padded if keys is None else {key: padded[key] for key in keys}
            rows = list(zip(*padded_.values()))
            empty_rows = [[' '*chars[key]] for key in padded_]
            ns = [max([len(v) for v in row]) for row in rows]
            if len(rowsep) == 1:
                sep_rows = [[rowsep*chars[key]] for key in padded_]
                res = ['\n'.join([colsep.join(v) for v in zip(*[r + e * (n-len(r)) + s for r,e,s in zip(row, empty_rows, sep_rows)])]) 
                       for row,n in zip(rows, ns)]
            else:
                res = ['\n'.join([colsep.join(v) for v in zip(*[r+e*(n-len(r)) for r, e in zip(row, empty_rows)])]) for row, n in zip(rows, ns)]
            if mid and cat and len(self)>2*cat:
                n = (len(res)+1)//2
                res = res[:n] + [mid] + res[n:]
            if rowsep == 'header':
                res = res[:1] + ['+'.join(['-'*chars[key] for key in padded_])] + res[1:]
            if header:
                res = [header] + res
            if footer:
                res = res + [footer]
            return '\n'.join(res)
        if max_width is None or max_width<=0:
            return text(padded)
        else:
            keys = list(padded.keys())
            res = []
            while len(keys)>0:
                i = 0
                width = 0
                while width<max_width and i<len(keys):
                    width+= chars[keys[i]]
                    i+=1
                res.append(text(padded, keys[:i]))
                keys = keys[i:]
            return '\n\n'.join(res)

    def sort(self, *by):
        """
        Sorts the table either using a key, list of keys or functions of members
        
        
        :Example:
        -------
        >>> import numpy as np
        >>> self = dictable(a = [_ for _ in 'abracadabra'], b=range(11), c = range(0,33,3))
        >>> self.d = list(np.array(self.c) % 11)
        >>> res = self.sort('a', 'd')
        >>> assert list(res.c) == list(range(11))
        
                
        >>> d = dictable(a = ['a', 1, 'c', 0, 'b', 2]).sort('a')        
        >>> res = d.sort('a','c')
        >>> print(res)
        >>> assert ''.join(res.a) == 'aaaaabbcdrr' and list(res.c) == [0,4,8,9,10] + [2,3] + [1] + [7] + [5,6]
        
        >>> d = d.sort(lambda b: b*3 % 11) ## sorting again by c but using a function
        >>> assert list(d.c) == list(range(11))
        """
        by = as_list(by)
        if len(self) == 0 or len(by) == 0:
            return self.copy()
        keys = self[as_tuple(by)]
        keys2id = list(zip(keys, range(len(self))))
        _, rows = zip(*sort(keys2id))
        return type(self)({key: [value[i] for i in rows] for key, value in self.items()})

    def __add__(self, other):
        if other in (None, 0):
            return self
        return self.concat(self, other)                
    
    __radd__ = __add__
    
    def _listby(self, by):
        keys = self[by]
        keys2id = sort(list(zip(keys, range(len(self)))))
        prev = None
        res = []
        row = []
        for key, i in keys2id:
            if len(row) == 0 or key==prev:
                row.append(i)
            else:
                res.append((prev, row))
                row = [i]
            prev = key 
        res.append((prev, row))
        return zip(*res)

    def listby(self,*by):
        if len(self) == 0:
            return self.copy()
        if len(by) == 0:
            by = self.keys()
        by = as_tuple(by)
        if len(by) == 0:
            return dictable({key: [value] for key, value in dict(self).items()})
        xs, ids = self._listby(by)
        rtn = dictable(xs, by)
        rtn.update({k: [[self[k][i] for i in y] for y in ids] for k in self.keys() if k not in by})
        return rtn
    
    @classmethod
    def read_excel(cls, io, sheet_name = None, floats = None, ints = None, dates = None, no_none = None):
        if is_str(io):
            io = io.lower()
            if '!' in io and sheet_name is None: ## handling input like: ExcelFile.xlsx!SheetName
                io, sheet_name = io.split('!')
            if not '.xls' in io:
                io = io + '.xlsx'
            io = pd.ExcelFile(io)
        if sheet_name is None:
            sheet_name = 0
        df = io.parse(sheet_name)
        res = cls(df)
        res = res.relabel(lambda v: v.strip().replace(' ', '_'))
        res = res.do(nan2none)
        if floats:
            res = res.do(try_back(as_float), floats)
        if ints:
            res = res.do(try_back(int), ints)
        if dates:
            res = res.do(try_back(ndt), dates)
        if no_none:
            for col in as_list(no_none):
                res = res.exc(**{col : None})
        return res
    
    def unlist(self):
        """
        undoes listby...
        
        :Example:
        -------
        >>> x = dictable(a = [1,2,3,4], b= [1,0,1,0])
        >>> x.listby('b')
        
        dictable[2 x 2]
        b|a     
        0|[2, 4]
        1|[1, 3]
        
        
        >>> assert x.listby('b').unlist().sort('a') == x

        :Returns:
        -------
        dictable
            a dictable where all rows with list in them have been 'expanded'.

        """
        return self.concat([row for row in self]) if len(self) else self
    
    def groupby(self,*by, grp = 'grp'):
        """
        Similar to pandas groupby but returns a dictable of dictables with a new column 'grp' 
        
        :Example:
        -------
        >>> x = dictable(a = [1,2,3,4], b= [1,0,1,0])
        >>> res = x.groupby('b')
        >>> assert res.keys() == ['b', 'grp']
        >>> assert is_dictable(res[0].grp) and res[0].grp.keys() == ['a']

        :Parameters:
        ----------------
        *by : str or list of strings
        
            gr.
        grp : str, optional
            The name of the column for the dictables per each key. The default is 'grp'.

        :Returns:
        -------
        dictable
            A dictable containing the original keys and a dictable per unique key.

        """
        if len(self) == 0:
            return self.copy()
        if len(by) == 0:
            by = self.keys()
        by = as_tuple(by)
        if len(by) == 0:
            raise ValueError('cannot groupby on no keys, left with original dictable')
        elif len(by) == len(self.keys()):
            raise ValueError('cannot groupby on all keys... nothing left to group')
        xs,ys = self._listby(by)
        rtn = dictable(xs, by)
        rtn[grp] = [dictable({k: [self[k][i] for i in y] for k in self.keys() if k not in by}) for y in ys]
        return rtn

    def ungroup(self, grp = 'grp'):
        """
        Undoes groupby

        :Example:
        -------
        >>> x = dictable(a = [1,2,3,4], b= [1,0,1,0])
        >>> self = x.groupby('b')
        
        :Parameters:
        ----------------
        grp : str, optional
            column name where dictables are. The default is 'grp'.

        :Returns:
        -------
        dictable.

        """
        return self.concat([row.pop(grp)(**row.do(lambda v: [v])) for row in self])
        
    
    def join(self, other, lcols = None, rcols = None, mode = None):
        """
        
        Performs either an inner join or a cross join between two dictables

        :Example: inner join
        -------------------------------
        >>> from pyg import *
        >>> x = dictable(a = ['a','b','c','a']) 
        >>> y = dictable(a = ['a','y','z'])
        >>> assert x.join(y) == dictable(a = ['a', 'a'])

        :Example: outer join
        -------------------------------
        >>> from pyg import *
        >>> x = dictable(a = ['a','b']) 
        >>> y = dictable(b = ['x','y'])
        >>> assert x.join(y) == dictable(a = ['a', 'a', 'b', 'b'], b = ['x', 'y', 'x', 'y'])

        """
        _lcols = lcols
        if not isinstance(other, dictable):
            other = dictable(other)
        if lcols is None:
            lcols = self.keys() & other.keys()
        if rcols is None:
            rcols = lcols
        lcols = as_tuple(lcols); rcols = as_tuple(rcols)
        if len(lcols)!=len(rcols):
            raise ValueError('cannot inner join when cols on either side mismatch in length %s vs %s'%(lcols, rcols))
        elif _lcols is None:
            if len(lcols) == 0:
                print('outer joining')
            else:
                print('inner joining on %s'%list(lcols))
        cols = []
        for lcol, rcol in zip(lcols, rcols):
            if is_str(lcol):
                cols.append(lcol)
            elif is_str(rcol):
                cols.append(rcol)
            else:
                raise ValueError('Cannot use a formula to inner join on both left and right %s %s'%(lcol, rcol))
        lkeys = self.keys() - cols
        rkeys = other.keys() - cols
        jkeys = lkeys & rkeys
        lkeys = lkeys - jkeys
        rkeys = rkeys - jkeys
        
        if len(cols):
            lxs, lids = self._listby(lcols)
            rxs, rids = other._listby(rcols)
            ls = len(lxs)
            rs = len(rxs)
            l = 0
            r = 0
            res = []
            while l<ls and r<rs:
                while l<ls and r<rs and cmp(lxs[l],rxs[r]) == -1:
                    l+=1
                while l<ls and r<rs and cmp(lxs[l],rxs[r]) == 1:
                    r+=1
                if l<ls and r<rs and lxs[l] == rxs[r]:
                    res.append((lxs[l], lids[l], rids[r]))
                    r+=1
                    l+=1
            if len(res) == 0:
                return dictable([], cols + lkeys + rkeys + jkeys)
            xs, lids, rids = zip(*res)
            ns = [len(l)*len(r) for l, r in zip(lids, rids)]
            rtn = dictable(sum([[x]*n for x,n in zip(xs,ns)], []), cols)
        else:
            rtn = dictable()
            lids = [range(len(self))]; rids = [range(len(other))]
        for k in lkeys:
            v= self[k]
            rtn[k] = sum([[v[l] for l in lid for r in rid] for lid, rid in zip(lids, rids)], [])
        for k in rkeys:
            v= other[k]
            rtn[k] = sum([[v[r] for l in lid for r in rid] for lid, rid in zip(lids, rids)], [])            
        for k in jkeys:
            if (is_str(mode) and mode[0].lower() == 'l') or mode == 0:
                v = self[k]
                rtn[k] = sum([[v[l] for l in lid for r in rid] for lid, rid in zip(lids, rids)], [])
            elif (is_str(mode) and mode[0].lower() == 'r') or mode == 1:
                v = other[k]
                rtn[k] = sum([[v[r] for l in lid for r in rid] for lid, rid in zip(lids, rids)], [])
            elif callable(mode):
                lv = self[k]
                rv = other[k]
                rtn[k] = sum([[mode(lv[l], rv[r]) for l in lid for r in rid] for lid, rid in zip(lids, rids)], [])            
            else:                
                lv = self[k]
                rv = other[k]
                rtn[k] = sum([[(lv[l], rv[r]) for l in lid for r in rid] for lid, rid in zip(lids, rids)], [])            
        return rtn
    
    def xor(self, other, lcols = None, rcols = None, mode = 'l'):
        """
        returns what is in lhs but NOT in rhs (or vice versa if mode = 'r'). Together with inner joining, can be used as left/right join

        :Examples:
        --------------
        >>> from pyg import *
        >>> self = dictable(a = [1,2,3,4])
        >>> other = dictable(a = [1,2,3,5])
        >>> assert self.xor(other) == dictable(a = 4) # this is in lhs but not in rhs
        >>> assert self.xor(other, lcols = lambda a: a * 2, rcols = 'a') == dictable(a = [2,3,4]) # fit can be done using formulae rather than actual columns

        The XOR functionality can be performed using quotient (divide):
        >>> assert lhs/rhs == dictable(a = 4)
        >>> assert rhs/lhs == dictable(a = 5)

        >>> rhs = dictable(a = [1,2], b = [3,4])
        >>> left_join_can_be_done_simply_as = lhs * rhs + lhs/rhs


        :Parameters:
        ----------------
        other : dictable (or something that can be turned to one)
            what we exclude with.
        lcols : str/list of strs, optional
            the left columns/formulae on which we match. The default is None.
        rcols : str/list of strs, optional
            the right columns/formulae on which we match. The default is None.
        mode : string, optional
            When set to 'r', performs xor the other way. The default is 'l'.

        :Returns:
        -------
        dictable
            a dictable containing what is in self but not in ther other dictable.

        """
        if not isinstance(other, dictable):
            other = dictable(other)
        if lcols is None:
            lcols = self.keys() & other.keys()
        if rcols is None:
            rcols = lcols
        lcols = as_tuple(lcols); rcols = as_tuple(rcols)
        if len(lcols)!=len(rcols):
            raise ValueError('cannot xor-join when cols on either side mismatch in length %s vs %s'%(lcols, rcols))
        if len(lcols) == 0:
            return self.copy()
        lxs, lids = self._listby(lcols)
        rxs, rids = other._listby(rcols)
        mode = 1 if (is_str(mode) and mode[0].lower() == 'r') or mode == 1 else 0
        ls = len(lxs)
        rs = len(rxs)
        l = 0
        r = 0
        res = []
        while l<ls and r<rs:
            while l<ls and r<rs and cmp(lxs[l],rxs[r]) == -1:
                if mode == 0:
                    res.append(lids[l])
                l+=1
            while l<ls and r<rs and cmp(lxs[l],rxs[r]) == 1:
                if mode == 1:
                    res.append(rids[r])                    
                r+=1
            if l<ls and r<rs and lxs[l] == rxs[r]:
                r+=1
                l+=1
        if mode == 0:
            if l<ls:
                res.extend(lids[l:])
            return self[sum(res, [])]
        else:
            if r<rs:
                res.extend(rids[r:])
            return other[sum(res, [])]

    __mul__ = join
    __truediv__ = xor
    __div__ = xor
    
    def xyz(self, x, y, z, agg = None):
        """
        
        pivot table functionality.
        
        :Parameters:
        ----------------
        x : str/list of str
            unique keys per each row
        y : str
            unique key per each column
        z : str/callable
            A column in the table or an evaluated quantity per each row
        agg : None/callable or list of callables, optional
            Each (x,y) cell can potentially contain multiple z values. so if agg = None, a list is returned
            If you want the data aggregated in any way, then supply an aggregating function(s)

        :Returns:
        -------
        A dictable which is a pivot table of the original data
        

        :Example:
        -------
        >>> from pyg import *
        >>> timetable_as_list = dictable(x = [1,2,3]) * dictable(y = [1,2,3]) 
        >>> timetable = timetable_as_list.xyz('x','y',lambda x, y: x * y)
        >>> assert timetable = dictable(x = [1,2,3], )

        :Example:
        -------
        >>> self = dictable(x = [1,2,3]) * dictable(y = [1,2,3]) 
        >>> x = 'x'; y = 'y'; z = lambda x, y: x * y
        >>> self.exc(lambda x, y: x+y==5).xyz(x,y,z, len)
        
        """
        if not is_strs(x):
            raise ValueError('x must be columns %s'%x)
        agg = as_list(agg)
        x = as_tuple(x)
        xykeys = x + as_tuple(y)
        xys, ids = self._listby(xykeys)
        zs = self[z]
        y_ = y if is_str(y) else '_columns'
        rs = dictable(xys, x + (y_,))        
        ys = rs[as_list(y_)].listby(y_)
        y2id = dict(zip(ys[y_], range(len(ys))))
        xs, yids = rs._listby(x)
        res = [[None for _ in range(len(ys))] for _ in range(len(xs))]
        for i in range(len(xs)):
            for j in yids[i]:
                xy = xys[j]
                k = y2id[xy[-1]]
                value = [zs[id_] for id_ in ids[j]]
                if agg:
                    for a in agg:
                        value = a(value)
                res[i][k] = value
        dx = dictable(xs, x)
        dy = dictable(res, y2id.keys())
        dx.update(dy)
        return dx
    
    pivot = xyz
    
    def unpivot(self, x, y, z):
        """
        undoes self.xyz / self.pivot
        
        :Example:
        -------
        >>> from pyg import *
        >>> orig = (dictable(x = [1,2,3,4]) * dict(y = [1,2,3,4,5]))(z = lambda x, y: x*y)
        >>> pivot = orig.xyz('x', 'y', 'z', last)
        >>> unpivot = pivot.unpivot('x','y','z').do(int, 'y') # the conversion to column names mean y is now string... so we convert back to int
        >>> assert orig == unpivot

        :Parameters:
        ----------------
        x : str/list of strings
            list of keys in the pivot table.
        y : str
            name of the columns that wil be used for the values that are currently column headers.
        z : str
            name of the column that describes the data currently within the pivot table.

        :Returns:
        -------
        dictable

        """
        xcols = as_list(x)
        if isinstance(y, dict) and len(y) == 1:
            y, ycols = list(y.items())[0]
        else:
            ycols = self.keys() - x
        ycols = as_tuple(ycols)
        n = len(ycols)

        res = dictable({k: sum([[row[k]]*n for row in self], []) for k in xcols})
        res[y] = ycols * len(self)
        res[z] = sum([[row[ycol] for ycol in ycols] for row in self], [])
        return res
    

            

def is_dictable(value):
    return isinstance(value, dictable)




from functools import reduce
from pyg_base._decorators import wrapper
from pyg_base._types import is_str

__all__ = ['reducer', 'reducing']

def reducer(function, sequence, default = None):
    """
    reduce adds stuff to zero by defaults. This is not needed.

    :Parameters:
    ----------------
    function : callable
        binary function.
    sequence : iterable
        list of inputs to be applied iteratively to reduce.
    default : TYPE, optional
        A default value to be returned with an empty sequence

    :Example:
    -------
    >>> from operator import add, mul
    >>> from functools import reduce
    >>> import pytest
    
    >>> assert reducer(add, [1,2,3,4]) == 10
    >>> assert reducer(mul, [1,2,3,4]) == 24
    >>> assert reducer(add, [1]) == 1
    
    >>> assert reducer(add, []) is None
    >>> with pytest.raises(TypeError):
    >>>     reduce(add, [])
    """
    sequence = list(sequence)
    if len(sequence) == 0:
        return default
    elif len(sequence) == 1:
        return sequence[0]
    else:
        return reduce(function, sequence[1:], sequence[0])

    
class reducing(wrapper):
    """
    Makes a bivariate-function being able to act on a sequence of elements using reduction
    
    :Example:
    -------
    >>> from operator import mul
    >>> assert reducing(mul)([1,2,3,4]) == 24    
    >>> assert reducing(mul)(6,4) == 24    

    Since a.join(b).join(c).join(d) is also very common, we provide a simple interface for that:

    :Example: chaining
    -----------------------------
    >>> assert reducing('__add__')([1,2,3,4]) == 10
    >>> assert reducing('__add__')(6,4) == 10
    
    d = dictable(a = [1,2,3,5,4])
    reducing('inc')(d, dict(a=1))
    
    """
    def wrapped(self, lhs, rhs = None, default = None, **kwargs):
        if is_str(self.function):
            func = lambda lhs, rhs: getattr(lhs, self.function)(rhs, **kwargs)
        else:
            if len(kwargs):
                func = lambda lhs, rhs: self.function(lhs, rhs, **kwargs)
            else:
                func = self.function
        if rhs is None:
            return reducer(func, lhs, default)
        else:
            return func(lhs, rhs)




import numpy as np
from pyg_base._logger import logger
from pyg_base._inspect import getargspec, getargs
from pyg_base._dictattr import dictattr
from copy import copy
import datetime


__all__ = ['wrapper', 'try_back', 'try_nan', 'try_none', 'try_zero', 'try_false', 'try_true', 'try_list', 'timer']
_function = 'function'

class wrapper(dictattr):
    """
    A base class for all decorators. It is similar to functools.wraps but better. See below why wrapt cannot be used...
    You basically need to define the wrapped method and everything else is handled for you.
    - You can then use it either directly to decorate functions
    - Or use it to create parameterized decorators
    - the __name__, __wrapped__, __doc__ and the getargspec will all be taken care of.
    
    :Example:
    -------
    >>> class and_add(wrapper):
    >>>     def wrapped(self, *args, **kwargs):
    >>>         return self.function(*args, **kwargs) + self.add ## note that we are assuming self.add exists
    
    >>> @and_add(add = 3) ## create a decorator and decorate the function
    >>> def f(a,b):
    >>>     return a+b
        
    >>> assert f.add == 3
    >>> assert f(1,2) == 6
    

    Alternatively you can also use it this directly:
        
    >>> def f(a,b):
    >>>     return a+b
    >>> 
    >>> assert and_add(f, add = 3)(1,2) == 6

    
    :Example: Explicit parameter construction
    -----------------------------------------
    
    You can make the init more explict, also adding defaults for the parameters:

    >>> class and_add_version_2(wrapper):
    >>>     def __init__(self, function = None, add = 3):
    >>>         super(and_add, self).__init__(function = function, add = add)
    >>>     def wrapped(self, *args, **kwargs):
    >>>         return self.function(*args, **kwargs) + self.add

    >>> @and_add_version_2
    >>> def f(a,b):
    >>>     return a+b
    >>> assert f(1,2) == 6
        

    :Example: No recursion
    ------------------
    The decorator is designed to have a single instance of a specific wrapper
    
    >>> f = lambda a, b: a+b
    >>> assert and_add(and_add(f)) == and_add(f)

    This holds even for multiple levels of wrapping:

    >>> x = try_none(and_add(f))
    >>> y = try_none(and_add(x))
    >>> assert x == y        
    >>> assert x(1, 'no can add') is None        

    :Example: wrapper vs wrapt
    --------------------------
    wrapt (wrapt.readthedocs.io) is an awesome wrapping tool. If you have static library functions, none is better.
    The problem we face is that wrapt is too good in pretending the wrapped up object is the same as original function:
        
    >>> import wrapt    
    >>> def add_value(value):
    >>>     @wrapt.decorator
    >>>     def wrapper(wrapped, instance, args, kwargs):
    >>>         return wrapped(*args, **kwargs) + value
    >>>     return wrapper

    >>> def f(x,y):
    >>>     return x*y

    >>> add_three = add_value(value = 3)(f)
    >>> add_four = add_value(value = 4)(f)
    >>> assert add_four(3,4) == 16 and add_three(3,4) == 15

    >>> ## but here is the problem:
    >>> assert encode(add_three) == encode(add_four) == encode(f)
    
    So if we ever encode the function and send it across json/Mongo, the wrapping is lost and the user when she receives it cannot use it

    >>> class add_value(wrapper):
    >>>     def wrapped(self, *args, **kwargs):
    >>>         return self.function(*args, **kwargs) + self.value

    >>> add_three = add_value(value = 3)(f)
    >>> add_four = add_value(value = 4)(f)
    >>> encode(add_three)
    >>> {'value': 3, 'function': '{"py/function": "__main__.f"}', '_obj': '{"py/type": "__main__.add_value"}'}
    >>> encode(add_three)
    >>> {'value': 4, 'function': '{"py/function": "__main__.f"}', '_obj': '{"py/type": "__main__.add_value"}'}
 
    """    
    def __init__(self, function = None, *args, **kwargs):
        function = copy(function)
        if type(function) == type(self):
            kw = function._kwargs
            kw.update(kwargs)
            function = function.function
        else:
            kw = kwargs
        f = function
        while isinstance(f, wrapper):
            if type(f.function) == type(self):
                kw = f.function._kwargs
                kw.update(kwargs)
                f[_function] = f.function.function
            else:
                f = f.function

        super(wrapper, self).__init__(*args, **kw)
        self['function'] = function
        bad_keys = [key for key in kwargs if key.startswith('_')]
        if len(bad_keys):
            raise ValueError('Cannot wrap _hidden parameters %s'%bad_keys)
        for attr in ['doc']:
            attr = '__%s__'%attr
            if hasattr(self.function, attr):
                setattr(self, attr, getattr(self.function, attr))

    @property
    def __name__(self):
        return getattr(self.function, '__name__', 'pyg_base.wrapper(unnamed)')
    
    @property
    def __wrapped__(self):
        return self.function

    @property
    def fullargspec(self):
        return getargspec(self.function)

    def __repr__(self):
        return '%s(%s)'%(self.__class__.__name__, dict(self))

    __str__ = __repr__ 

    @property
    def _kwargs(self):
        return {key: value for key, value in self.items() if key!='function'}

    def __call__(self, *args, **kwargs):
        if self.function is None and len(args) == 1 and len(kwargs) == 0:
            return type(self)(function = args[0], **self._kwargs)
        else:
            return getattr(self, 'wrapped', self.function)(*args, **kwargs)
    
class try_back(wrapper):
    """
    wraps a function to try an evaluation. If an exception is thrown, returns first argument

    :Example:
    --------------
    >>> f = lambda a: a[0]
    >>> assert try_back(f)('hello') == 'h' and try_back(f)(5) == 5
    """
    def __init__(self, function = None):
        super(try_back, self).__init__(function = function)
    def wrapped(self, *args, **kwargs):
        try:
            return self.function(*args, **kwargs)
        except Exception:
            return args[0] if len(args)>0 else kwargs[getargs(self.function)[0]]


class try_value(wrapper):
    """
    wraps a function to try an evaluation. If an exception is thrown, returns a cached argument
    
    :Parameters:
    ------------
    function callable
        The function we want to decorate
    value: 
        If the function fails, it will return value instead. Default is None
    verbose: bool
        If set to True, the logger will warn with the error message.

    There are various convenience functions with specific values
    try_zero, try_false, try_true, try_nan and try_none will all return specific values if function fails.
    
    :Example:
    --------------
    >>> from pyg import *
    >>> f = lambda a: a[0]
    >>> assert try_none(f)(4) is None
    >>> assert try_none(f, 'failed')(4) == 'failed'

    
    """
    def __init__(self, function = None, value = None, verbose = None):
        super(try_value, self).__init__(function = function, value = value, verbose = verbose)
    def wrapped(self, *args, **kwargs):
        try: 
            return self.function(*args, **kwargs)
        except Exception as e:
            if self.verbose:
                logger.warning('WARN: %s' % e)
            return copy(self.value)

try_nan = try_value(value = np.nan)
try_zero = try_value(value = 0)
try_none = try_value
try_true = try_value(value = True)
try_false = try_value(value = False)
try_list = try_value(value = [])


def _str(value):
    """
    returns a short string:
    >>> _str([1,2,3])
    :Parameters:
    ----------------
    value : TYPE
        DESCRIPTION.

    :Returns:
    -------
    TYPE
        DESCRIPTION.

    """
    if isinstance(value, (int, str, float, bool, datetime.datetime, datetime.date)):
        return str(value) 
    elif hasattr(value, '__len__'):
        return '%s[%i]'%(type(value), len(value))
    else:
        return str(type(value))
        
_txt = 'TIMER:%r args:[%r, %r] (%i runs) took %s sec'    
class timer(wrapper):
    """
    timer is similar to timeit but rather than execution of a Python statement, 
    timer wraps a function to make it log its evaluation time before returning output
    
    :Parameters:
    ------------
    function: callable
        The function to be wraooed 
    
    n: int, optional
        Number of times the function is to be evaluated. Default is 1
    
    time: bool, optional
        If set to True, function will return the TIME it took to evaluate rather than the original function output.


    :Example:
    ---------
    >>> from pyg import *; import datetime
    >>> f = lambda a, b: a+b
    >>> evaluate_100 = timer(f, n = 100, time = True)(1,2)
    >>> evaluate_10000 = timer(f, n = 10000, time = True)(1,2)
    >>> assert evaluate_10000> evaluate_100
    >>> assert isinstance(evaluation_time, datetime.timedelta)
    """
    
    def __init__(self, function, n = 1, time = False):
        super(timer, self).__init__(function = function, n = n, time = time)

    def wrapped(self, *args, **kwargs):
        t0 = datetime.datetime.now()
        for _ in range(self.n):
            res = self.function(*args, **kwargs)
        t1 = datetime.datetime.now()
        time = t1 - t0
        logger.info(_txt%(getattr(self.function,'__name__', ''), 
                            [_str(a) for a in args],
                            ['%s=%s'%(key, _str(value)) for key, value in kwargs.items()],
                            self.n,
                            time
                            ))        
        return time if self.time else res

class kwargs_support(wrapper):
    """
    Extends a function to support **kwargs inputs
    
    :Example:
    ---------
    >>> from pyg import *
    >>> @kwargs_support
    >>> def f(a,b):
    >>>     return a+b
    
    >>> assert f(1,2, what_is_this = 3, not_used = 4, ignore_this_too = 5) == 3
    
    """
    def __init__(self, function = None):
        super(kwargs_support, self).__init__(function = function)
    @property
    def _args(self):
        return getargs(self.function)
        
    def wrapped(self, *args, **kwargs):
        _args = self._args
        kwargs = {key : value for key, value in kwargs.items() if key in _args}
        return self.function(*args, **kwargs)
 

        


from pyg_base._as_list import as_list, is_rng
from pyg_base._eq import eq
from pyg_base._types import is_str
from pyg_base._ulist import ulist

__all__ = ['dictattr', 'relabel']

class dictattr(dict):
    '''
    A simple dict with extended member manipulation
        1) access using d.key
        2) access multiple elements using d[key1, key2]
    
    
    :Example: members access
    ------------------------
    >>> from pyg import *
    >>> d = dictattr(a = 1, b = 2, c = 3)
    >>> assert isinstance(d, dict)
    >>> assert d.a == 1
    >>> assert d['a','b'] == [1,2]
    >>> assert d[['a','b']] == dictattr(a = 1, b = 2)


    In addition, it has extended key selection/subsetting
    
    :Example: subsetting
    -------------------
    >>> d = dictattr(a = 1, b = 2, c = 3)
    >>> assert d - 'a' == dictattr(b = 2, c = 3)
    >>> assert d & ['b', 'c', 'not in keys'] == dictattr(b = 2, c = 3)


    dictattr supports not in-place 'update':

    :Example: updating via adding another dict
    ------------------------------------------
    >>> d = dictattr(a = 1, b = 2) + dict(b = 'replacing old value', c = 'new key')
    >>> assert d == dictattr(a = 1, b = 'replacing old value', c = 'new key')

    '''
    def __sub__(self, key, copy = True):
        """
        deletes an item but does not throw an exception if not there
        dictattr uses subtraction to remove key(s)

        :Returns:
        -------
        updated dictattr
        
        
        :Example:
        -------
        >>> from pyg import *
        >>> d = dictattr(a = 1, b = 2, c = 3)
        >>> assert d - ['b','c'] == dictattr(a = 1)
        >>> assert d - 'c' == dictattr(a = 1, b = 2)
        >>> assert d - 'key not there' == d
        >>> #commutative
        >>> assert (d - 'c').keys() == d.keys() - 'c'
        """
        res = self.copy() if copy else self
        if isinstance(key, tuple):
            branch = res
            for k in key[:-1]:
                if k in branch:
                    branch = branch[k]
                else:
                    return res
            if key[-1] in branch:
                del branch[key[-1]]
        elif isinstance(key, list):
            for k in key:
                res = res.__sub__(k, copy = False)
        else:
            if key in res:
                del res[key]               
        return res

    def __and__(self, other):
        """
        dictattr uses & as a set operator for key filtering

        :Returns:
        -------
        updated dictattr
        
        :Example:
        -------
        >>> from pyg import *
        >>> d = dictattr(a = 1, b = 2, c = 3)
        >>> assert d & ['a', 'b', 'not_there'] == dictattr(a = 1, b = 2)
        >>> #commutative
        >>> assert (d & ['a', 'b', 'x']).keys() == d.keys() & ['a', 'b', 'x']
        """
        other = set(as_list(other))
        return type(self)(**{key : value for key, value in self.items() if key in set(self.keys()) & other}) 

    def __add__(self, other):
        """
        dictattr uses add as a copy + update. Similar to the latest python |=        
        
        :Example:
        ---------
        >>> from pyg import *
        >>> d = dictattr(a = 1, b = 2)
        >>> assert d + dict(b = 3, c = 5) == dictattr(a = 1, b = 3, c = 5)
        
        :Parameters:
        ------------
        other: dict
            a dict used to update current dict.
        
        """
        res = self.copy()
        res.update(other)
        return res
    
    def __truediv__(self, other):
        if type(other) == type:
            return type(self)({key: value for key, value in self.items() if not isinstance(value, other)})                        
        elif callable(other):
            return type(self)({key: value for key, value in self.items() if not other(value)})            
        else:
            return type(self)({key: value for key, value in self.items() if not eq(value, other)})
    
    def __dir__(self):
        return list(self.keys()) + super(dictattr, self).__dir__()

    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError as e: ## must be raised as an AttributeError so that getattr() works
            if attr.startswith('_'):
                return getattr(dict(self), attr)
            else:
                raise AttributeError(e)
    
    def __setattr__(self, attr, value):
        if attr.startswith('_'):
            super(dictattr, self).__setattr__(attr, value)
        else:
            self[attr] = value
    
    def __delattr__(self, attr):
        try:
            del self[attr]
        except KeyError as e:
            if attr.startswith('_'):
                super(dictattr, self).__delattr__(attr)
            else:
                raise AttributeError(str(e))

    def __delitem__(self, key):
        if isinstance(key, tuple):
            res = self
            for k in key[:-1]:
                res = res[k]
            del res[key[-1]]
        elif isinstance(key, list):
            for k in key:
                del self[k]
        else:
            try:
                super(dictattr, self).__delitem__(key)               
            except KeyError as e:
                if is_str(key) and '.' in key:
                    res = self
                    keys = key.split('.')
                    for k in keys[:-1]:
                        res = res[k]
                    del res[keys[-1]]
                else:
                    raise KeyError(e)
    
    def __getitem__(self, value):
        if isinstance(value, tuple):
            return [self[v] for v in value]
        elif is_rng(value):
            return type(self)(**{k : self[k] for k in value})
        res = self
        if value in res or not is_str(value):
            return super(dictattr, self).__getitem__(value)
        else:
            for k in value.split('.'):
                res = dict(res)[k]
            return res
    
    def copy(self):
        return type(self)(**self)
    
    def __getstate__(self):
        return dict(self)
    
    def __setstate__(self, state):
        self.update(state)
    
    def __dict__(self):
        return dict(self)
    
    def keys(self):
        """
        dictattr returns an actual list rather than a generator. 
        Further, this recognises that the keys are necessarily unique so it returns a ulist which is also a set

        :Returns:
        -------
        ulist
            list of keys of dictattr.

        :Example:
        -------
        >>> from pyg import *
        >>> d = dictattr(a = 1, b = 2)
        >>> assert d.keys() == ulist(['a', 'b'])
        >>> assert d.keys() & ['a', 'c', 'd'] == ['a']
        """
        return ulist(super(dictattr, self).keys(), unique = True)
    
    def values(self):
        return list(super(dictattr, self).values())
    
    def relabel(self, *args, **relabels):
        """
        easy relabel/rename of keys

        :Parameters:
        ----------------
        *args : str or callable
            - a string ending/starting with _ will trigger a prefix/suffix to all keys
            - callable function will be applied to the keys to update them
            
        **relabels : strings
            individual relabeling of keys

        :Returns:
        -------
        dictattr
            new dict with renamed keys.
        
        :Example: suffix/prefix
        ------------------------
        >>> from pyg import *
        >>> d = dictattr(a = 1, b = 2, c = 3)
        >>> assert d.relabel('x_') == dictattr(x_a = 1, x_b = 2, x_c = 3) # prefixing
        >>> assert d.relabel('_x') == dictattr(a_x = 1, b_x = 2, c_x = 3) # suffixing

        :Example: callable 
        -------------------
        >>> assert d.rename(upper) == dictattr(A = 1, B = 2, C = 3)

        :Example: individual relabelling
        -------------------
        >>> assert d.rename(a = 'A') == dictattr(A = 1, b = 2, c = 3)
        >>> assert d.rename(['A', 'B', 'C']) == d.relabel(upper)
        """
        keys = relabel(list(self.keys()), *args, **relabels)
        return type(self)(**{keys.get(k,k) : v for k, v in self.items()})

    def rename(self, *args, **relabels):
        """
        Identical to relabel. See relabel for full docs
        """
        return self.relabel(*args, **relabels)
    
def relabel(keys, *args, **relabels):
    """
    returns a mapping from old keys to new keys
    
    :Example:
    --------------
    
    >>> assert relabel(['a', 'b', 'c'], a = 'A', b = 'B') == dict(a = 'A', b = 'B')
    >>> assert relabel(['a', 'b', 'c'], 'x_') == dict(a = 'x_a', b = 'x_b', c = 'x_c')
    >>> assert relabel(['a', 'b', 'c'], '_x') == dict(a = 'a_x', b = 'b_x', c = 'c_x')
    >>> assert relabel('a', lambda v: v * 2) == dict(a = 'aa')
    >>> assert relabel(['a', 'b'], lambda v: v * 2, b = 'other') == dict(a = 'aa', b = 'other')
    >>> assert relabel(['a', 'b', 'c'], 'A', 'B', 'C') == dict(a = 'A', b = 'B', c = 'C')
    >>> assert relabel(['a', 'b', 'c'], ['A', 'B', 'C']) == dict(a = 'A', b = 'B', c = 'C')
    
    :Parameters:
    ----------------
    keys : list
        a list of keys.
    *args : TYPE
        either 
        - a function to relabel keys.
        - a prefix/suffix
        - a list of new keys
    **relabels : TYPE
        DESCRIPTION.

    :Returns:
    -------
    res : dict
        a mapping from the old keys to the new keys. Note: result may not have all keys

    """
    keys = as_list(keys)
    args = as_list(args)
    res = {}
    if len(args) == 1:
        arg = args[0]
        if isinstance(arg , str):
            if arg.startswith('_'):
                res = {key : key + arg for key in keys}
            elif arg.endswith('_'):
                res = {key : arg + key for key in keys}
        elif callable(arg):
            res = {key : arg(key) for key in keys}
        elif isinstance(arg, dict):
            res.update(arg)
    elif len(args) == len(keys):
        res = {key : new_key  for key, new_key in zip(keys, args)}
    res.update(relabels)
    return res


from pyg_base._types import is_int, is_list

__all__ = ['rng', 'ulist']

def rng(*args):
    return list(range(*[a if is_int(a) else len(a) for a in args]))

class ulist(list):
    """
    A list whose members are unique. It has +/- operations overloaded while also supporting set opeations &/|
        
    :Example:
    ---------    
    >>> assert ulist([1,3,2,1]) == list([1,3,2])
    
    :Example: addition adds element(s)
    --------------------------------
    >>> assert ulist([1,3,2,1]) + 4  == list([1,3,2,4])
    >>> assert ulist([1,3,2,1]) + [4,1] == list([1,3,2,4])
    >>> assert ulist([1,3,2,1]) + [4,1,5] == list([1,3,2,4,5])

    :Example: subtraction removes element(s)
    -------------------------------------------
    >>> assert ulist([1,3,2,1]) - 1 == [3,2]
    >>> assert ulist([1,3,2,1]) - [1,3,4] == [2]

    :Example: set operations
    -------------------------------------------
    >>> assert ulist([1,3,2,1]) & 1 == [1]
    >>> assert ulist([1,3,2,1]) & [1,3,4] == [1,3]

    >>> assert ulist([1,3,2,1]) | 1 == [1,3,2]
    >>> assert ulist([1,3,2,1]) | 4 == [1,3,2,4]
    >>> assert ulist([1,3,2,1]) | [1,3,4] == [1,3,2,4]

    """
    def __init__(self, *args, unique = False):
        if unique:
            super(ulist, self).__init__(*args)
        else:
            orig = list(*args)
            super(ulist, self).__init__([v for _, v in sorted([(orig.index(u), u) for u in set(orig)])])

    def __add__(self, other):
        if is_list(other):
            return type(self)(super(ulist, self).__add__(other))
        elif other in self:
            return self.copy()
        else:
            return type(self)(super(ulist, self).__add__([other]))

    def copy(self):
        return type(self)(self, unique = True) 
    
    def __and__(self, other):
        if is_list(other):
            return type(self)([o for o in self if o in other])
        else:
            return type(self)([other], unique = True) if other in self else type(self)()
    
    __or__ = __add__
    
    def __sub__(self, other):
        if is_list(other):
            return type(self)([o for o in self if o not in other])
        elif other not in self:
            return self.copy()
        else:
            return type(self)([o for o in self if o not in [other]])
                



from pyg_base._dates import dt, DAY, TMIN, TMAX, dt_bump, is_period, is_bump, ymd
from pyg_base._types import is_int, is_str, is_ts, is_arr, is_pd, is_nan
from pyg_base._as_list import as_list 
from pyg_base._dict import Dict
from pyg_base._cache import cache
import datetime
from dateutil.rrule import rrule, MONTHLY, WEEKLY, DAILY, YEARLY, HOURLY, MINUTELY, SECONDLY, MO, TU, WE, TH, FR, SA, SU

import numpy as np
import re

weekdays = {0: MO, 1: TU, 2: WE, 3: TH, 4: FR, 5: SA, 6: SU}

HOUR = datetime.timedelta(hours = 1)
minute = datetime.timedelta(minutes = 1)
second = datetime.timedelta(seconds = 1)
microsecond = datetime.timedelta(microseconds = 1)
isoformat = re.compile('^[0-9]{4}-[0-9]{2}-[0-9]{2}T')

    
__all__ = ['date_range', 'drange', 'calendar', 'Calendar', 'clock', 'as_time']


def as_time(t = None):
    """
    parses t into a datetime.time object
    
    :Example:
    ---------
    >>> assert as_time('10:30:40') == datetime.time(10, 30, 40)
    >>> assert as_time('103040') == datetime.time(10, 30, 40)
    >>> assert as_time('10:30') == datetime.time(10, 30)
    >>> assert as_time('1030') == datetime.time(10, 30)
    >>> assert as_time('05') == datetime.time(5)
    >>> assert as_time(103040) == datetime.time(10, 30, 40)
    >>> assert as_time(13040) == datetime.time(1, 30, 40)
    >>> assert as_time(130) == datetime.time(1, 30)
    >>> assert as_time(datetime.time(1, 30)) == datetime.time(1, 30)
    >>> assert as_time(datetime.datetime(2000, 1, 1, 1, 30)) == datetime.time(1, 30)
    
    :Parameters:
    ----------
    t : str/int/datetime.time/datetime.datetime
        time of day

    :Returns:
    -------
    datetime.time

    """
    if t is None:
        t = dt()
    if is_str(t):
        hms = t.split(':')
        if len(hms) == 3:
            h,m,s = hms
            return datetime.time(int(h), int(m), int(s))
        elif len(hms) == 2:
            h,m = hms
            return datetime.time(int(h), int(m))
        else:
            t = int(t)
    if is_int(t):
        if t<100:
            return datetime.time(t)
        elif t<10000:
            h = t // 100
            m = t % 100
            return datetime.time(h, m)
        elif t<1000000:
            h = t // 10000
            t = t - h * 10000
            m = t // 100
            s = t % 100
            return datetime.time(h,m,s)
    elif isinstance(t, datetime.time):
        return t
    elif isinstance(t, datetime.datetime):
        return datetime.time(t.hour, t.minute, t.second)
    else:
        raise ValueError('cannot parse time %s'%t)

_LY = dict(b = DAILY, d = DAILY, w = WEEKLY, m = MONTHLY, q = MONTHLY, y = YEARLY, h = HOURLY, n = MINUTELY, s = SECONDLY)

def drange(t0 = None, t1 = None, bump = None):
    """
    A quick and happy wrapper for dateutil.rrule

    :Examples:
    --------------
    >>> drange(2000, 10, 1) # 10 days starting from dt(2000,1,1)
    >>> drange(2000, '10b', '1b') # weekdays between dt(2000,1,1) and dt(2000,1,17)
    >>> drange('-10b', 0, '1b') # business days since 10 bdays ago
    >>> drange('-10b', '10b', '1w') # starting 10b days ago, to 10b from now, counting in weekly jumps
    
    :Parameters:
    ----------------
    t0 : date, optional
        start date. The default is None.
    t1 : date, optional
        end date. The default is None.
    bump : timedelta, int, string, optional
        bump period. The default is None.

    :Returns:
    -------
    list of dates
    
    :Example:
    ---------
    >>> t0 = 2000; t1 = 1999
    >>> bump = '-1b'
    
    :Example:
    ---------
    >>> t0 = dt(2020); t1 = dt(2021); bump = datetime.timedelta(hours = 4)

    """
    t0, t1 = date_range(t0, t1)  
    if t0 == t1:
        return [t0]        
    if bump is None:
        bump = 1 if t0<t1 else -1
    if is_int(bump):
        if (t1-t0).days * bump <= 0:
            raise ValueError('cannot go from %s to %s in steps of %s'%(t0,t1,bump))
        freq = DAILY
        res = list(rrule(freq, interval = 1, dtstart = min(t0,t1), until = max(t0,t1))) # unfortunately does not actually work for negative bumps
        res = res[::-1] if bump<0 else res
        res = res[::abs(bump)] if abs(bump)>1 else res
        return res
    elif isinstance(bump, datetime.timedelta):
        t = t0
        res = []
        if t1>t0: 
            if t0 + bump <= t0:
                raise ValueError('cannot move forward from %s to %s using %s'%(t0, t1, bump))
            while t<=t1:
                res.append(t)
                t = t + bump
            return res
        elif t1<t0: 
            if t0 + bump >= t0:
                raise ValueError('cannot move back from %s to %s using %s'%(t0, t1, bump))
            while t>=t1:
                res.append(t)
                t = t + bump
            return res
        else:
            return [t0]
    elif is_period(bump):
        bump = bump.lower()
        period = bump[-1]
        interval = int(bump[:-1]) * dict(q = 3).get(period,1)
        if (t1-t0).days * interval < 0:
            raise ValueError('cannot go from %s to %s in steps of %s'%(t0,t1,bump))
        freq = _LY[period]
        if period == 'b':
            res = [t for t in rrule(freq, interval = 1, dtstart = min(t0,t1), until = max(t0,t1)) if t.weekday()<5]
            res = res[::-1] if interval<0 else res    
            res = res[::abs(interval)] if abs(interval)>1 else res
            return res            
        else:
            return list(rrule(freq, interval = interval, dtstart = t0, until = t1))


class _calendar():
    def dt_bump(self, t, bump):
        return dt_bump(t, bump)
    
    def date_range(self, t0 = None, t1 = None):
        """
        creates a date range boundaries from start/end points
    
        :Parameters:
        ----------------
        t0 : date or a date bump, optional
            start date. The default is None.
        t1 : date or a date bump, optional
            end date. The default is None.
    
        :Returns:
        -------
        sorted list of two dates
        
        
        :Examples:
        --------------
        >>> t = dt.today()
        >>> assert date_range() == [dt(1900), t]
        >>> assert date_range(-100, 100) == [dt_bump(t, -100), dt_bump(t,100)]
        >>> assert date_range(-100, '10b') == [dt_bump(t, -100), dt_bump(t,'10b')]
        >>> assert date_range(2000, '10b') == [dt(2000,1,1), dt(2000,1,17)]
    
        """
        t = dt(0)
        if t1 is None:
            if t0 is None:
                return [TMIN, t]
            elif is_bump(t0):
                return sorted([t, self.dt_bump(t, t0)])
            else:
                return sorted([t, dt(t0)])
        elif is_bump(t1):
            if t0 is None:
                return [t, self.dt_bump(t, t1)]
            elif is_bump(t0):
                return [self.dt_bump(t, t0), dt_bump(t, t1)]
            else:
                t0 = dt(t0)
                return [t0, self.dt_bump(t0, t1)]
        else:
            t1 = dt(t1)
            if t0 is None:
                return [t, t1]
            elif is_bump(t0):
                return [self.dt_bump(t1, t0), t1]
            else:
                return [dt(t0), t1]


_c  = _calendar()

def date_range(t0 = None, t1 = None):
    return _c.date_range(t0, t1)
        

"""
calendars is implemented as a factory of calendars indexed by key

We therefore choose to make any modifications to the calendar object in-place
"""
calendars = dict()

class Calendar(Dict, _calendar):
    """
    Calendar is 
        - a dict 
        - containing holiday dates 
        - implementing business day arithmetic

    Calendar is restricted to operate between cal.t0 and cal.t1 which default to TMIN = 1900 and TMAX = 2300

    Calendar does this by having two key members:
        - dt2int: a mapping from all business dates to their integer 'clock'
        - int2dt: a mapping from integer value to the date
    
    Since Calendar is an 'expensive' memory wise, we assign a key to the calendar and the Calendar is stored in the singleton calendars under this key
    
    :Example:
    -------
    >>> from pyg import *
    >>> holidays = dictable([[1,'2012-01-02','New Year Day',],
                            [2,'2012-01-16','Martin Luther King Jr. Day',],
                            [3,'2012-02-20','Presidents Day (Washingtons Birthday)',],
                            [4,'2012-05-28','Memorial Day',],
                            [5,'2012-07-04','Independence Day',],
                            [6,'2012-09-03','Labor Day',],
                            [7,'2012-10-08','Columbus Day',],
                            [8,'2012-11-12','Veterans Day',],
                            [9,'2012-11-22','Thanksgiving Day',],
                            [10,'2012-12-25','Christmas Day',],
                            [11,'2013-01-01','New Year Day',],
                            [12,'2013-01-21','Martin Luther King Jr. Day',],
                            [13,'2013-02-18','Presidents Day (Washingtons Birthday)',],
                            [14,'2013-05-27','Memorial Day',],
                            [15,'2013-07-04','Independence Day',],
                            [16,'2013-09-02','Labor Day',],
                            [17,'2013-10-14','Columbus Day',],
                            [18,'2013-11-11','Veterans Day',],
                            [19,'2013-11-28','Thanksgiving Day',],
                            [20,'2013-12-25','Christmas Day',],
                            [21,'2014-01-01','New Year Day',],
                            [22,'2014-01-20','Martin Luther King Jr. Day',],
                            [23,'2014-02-17','Presidents Day (Washingtons Birthday)',],
                            [24,'2014-05-26','Memorial Day',],
                            [25,'2014-07-04','Independence Day',],
                            [26,'2014-09-01','Labor Day',],
                            [27,'2014-10-13','Columbus Day',],
                            [28,'2014-11-11','Veterans Day',],
                            [29,'2014-11-27','Thanksgiving Day',],], ['i', 'date', 'name']).do(dt, 'date')

    
    >>> cal = calendar('US', holidays.date, t0 = 2012, t1 = 2015)
    >>> assert not cal.is_bday(dt(2013,9,2))        # Labor day

    >>> cached_calendar = calendar('US')
    >>> assert not cached_calendar.is_bday(dt(2013,9,2))   # Labor day

    >>> assert cal.adjust(dt(2013,9,2)) == dt(2013,9,3)
    >>> assert cal.drange(dt(2013,9,0), dt(2013,9,7), '1b') == [dt(2013,8,30), dt(2013,9,3), dt(2013,9,4), dt(2013,9,5), dt(2013,9,6),] ## skipped labour day and weekend prior
    
    >>> assert cal.bdays(dt(2013,9,0), dt(2013,9,7)) == 5
    """
    def __init__(self, key = None, holidays = None, weekend = None, t0 = None, t1 = None, adj = 'm'):
        """
        :Parameters:
        ----------------
        key : str, None, optional
            The key used to cache the calendar in singleton. The default is None.
        holidays : list of dates
            list of holidays
        weekend : list of int, optional
            What days are weekend 0=MON... 6 = SUN. The default is [5,6].
        t0 : datetime, optional
            calendar's first date. The default is TMIN = 1900.
        t1 : datetime, optional
            calendar's last date. The default is TMIN = 2300.
        adj : char, optional
            how non-bdays are adjusted. The default is 'm'.
            Three conventions are used:
                'f'ollowing : next date
                'p'revious: previous date
                'm'odified following: following date unless following date is in a different month, when we use previous.

        :Returns:
        -------
        Calendar object.

        """
        if isinstance(key, dict):
            super(Calendar, self).__init__(**key)
        if isinstance(key, Calendar):
            super(Calendar, self).__init__(key)
        else:
            if t1 is None:
                t1 = TMAX
            if t0 is None:
                t0 = TMIN
            t0, t1 = date_range(t0, t1)
            ## weekend are weekend days
            if weekend is None: 
                weekend = [5,6]
            else:
                weekend = as_list(weekend)
            holidays = as_list(holidays)
            holidays = dict(zip(holidays, holidays)) # we prefer to store holidays as a dict, as check of date in holidays is faster for hash
            super(Calendar, self).__init__(weekend = weekend, holidays = holidays,
                                           key = key, t0 = t0, t1 = t1, adj = adj)


    def _populate(self):
        """
        This is run once as it is quite expensive
        """
        if self.get('dt2int') is None or self.get('int2dt') is None:
            byweekday = tuple(v for k,v in weekdays.items() if k not in self.weekend)
            bdays = [date for date in rrule(DAILY, interval = 1, dtstart = self.t0, until = self.t1, byweekday = byweekday) if date not in self.holidays]
            self['dt2int'] = dict(zip(bdays, range(len(bdays))))        
            self['int2dt'] = dict(zip(range(len(bdays)), bdays))
        return self
        
    
    def __repr__(self):
        return 'calendar(%(key)s) from %(t0)s to %(t1)s'%self
    
    def is_holiday(self, date):
        return date.weekday() in self.weekend or ymd(date) in self.holidays
    
    def is_bday(self, date):
        return date.weekday() not in self.weekend and ymd(date) not in self.holidays

    def is_trading(self, date = None, day_start = 0, day_end = 235959):
        """
        calculates if we are within a trading session 

        :Parameters:
        ----------
        date : datetime, optional
            the time & date we want to check. The default is None (i.e. now)

        :Returns:
        -------
        bool:
            are we within a trading session
        """
        tod = as_time(date)
        day_start = as_time(day_start); day_end = as_time(day_end)
        day = ymd(date)
        if day_end >= day_start:
            return tod <= day_end and tod >= day_start and self.is_bday(day)
        else:
            if tod>=day_start and self.is_bday(day + DAY): # we are within tomorrow's session
                return True
            elif tod<=day_end and self.is_bday(day): # we are within today's session
                return True
            else:
                return False

    def trade_date(self, date = None, adj = None, day_start = 0, day_end = 235959):
        """
        This is very similar for adjust, but it also takes into account the time of the day.
        if day_start = 0 and day_end = 23:59:59 then this is exactly adjust.

        :Parameters:
        ----------
        date : datetime, optional
            date (with time). The default is None.
        adj : f/p, optional
            If date isn't within trading day, which direction to adjust to? The default is None.
            
        
        :Example:
        ---------
        >>> from pyg import *; import datetime

        >>> uk = calendar('UK', day_start = 8, day_end = 17)
        >>> assert uk.trade_date(dt(2021,2,9,5), 'f') == dt(2021, 2, 9)  # Tuesday morning rolls into Tuesday
        >>> assert uk.trade_date(dt(2021,2,9,5), 'p') == dt(2021, 2, 8)  # Tuesday morning back into Monday
        >>> assert uk.trade_date(dt(2021,2,7,5), 'f') == dt(2021, 2, 8)  # Sunday rolls into Monday
        >>> assert uk.trade_date(dt(2021,2,7,5), 'p') == dt(2021, 2, 5)  # Sunday rolls back to Friday

        >>> assert uk.trade_date(date = dt(2021,2,9,23), adj = 'f') == dt(2021, 2, 10)  # Tuesday eve rolls into Wed
        >>> assert uk.trade_date(date = dt(2021,2,9,23), adj = 'p') == dt(2021, 2, 9)  # Tuesday eve back into Tuesday
        >>> assert uk.trade_date(date = dt(2021,2,7,23), adj = 'f') == dt(2021, 2, 8)  # Sunday rolls into Monday
        >>> assert uk.trade_date(date = dt(2021,2,7,23), adj = 'p') == dt(2021, 2, 5)  # Sunday rolls back to Friday

        >>> assert uk.trade_date(date = dt(2021,2,9,12), adj = 'f') == dt(2021, 2, 9)  # Tuesday is Tuesday
        >>> assert uk.trade_date(date = dt(2021,2,9,12), adj = 'p') == dt(2021, 2, 9)  # Tuesday is Tuesday

        >>> au = calendar('AU', day_start = 2230, day_end = 1300)
        >>> assert au.trade_date(dt(2021,2,9,5), 'f') == dt(2021, 2, 9)  # Tuesday morning in session
        >>> assert au.trade_date(dt(2021,2,9,5), 'p') == dt(2021, 2, 9)  # Tuesday morning in session
        >>> assert au.trade_date(dt(2021,2,7,5), 'f') == dt(2021, 2, 8)  # Sunday rolls into Monday
        >>> assert au.trade_date(dt(2021,2,7,5), 'p') == dt(2021, 2, 5)  # Sunday rolls back to Friday

        >>> assert au.trade_date(date = dt(2021,2,9,23), adj = 'f') == dt(2021, 2, 10)  # Tuesday eve rolls into Wed
        >>> assert au.trade_date(date = dt(2021,2,9,23), adj = 'p') == dt(2021, 2, 10)  # Already in Wed
        >>> assert au.trade_date(date = dt(2021,2,7,23), adj = 'f') == dt(2021, 2, 8)  # Sunday rolls into Monday
        >>> assert au.trade_date(date = dt(2021,2,7,23), adj = 'p') == dt(2021, 2, 8)  # Already on Monday
        >>> assert au.trade_date(date = dt(2021,2,5,23), adj = 'f') == dt(2021, 2, 8)  # Friday afternoon rolls into Monday

        >>> assert au.trade_date(date = dt(2021,2,9,14), adj = 'f') == dt(2021, 2, 10)  # Tuesday is over, roll to Wed
        >>> assert au.trade_date(date = dt(2021,2,9,14), adj = 'p') == dt(2021, 2, 9)  # roll back to Tues
            
        """
        # first we split it into day and time of day
        tod = as_time(date)
        day = ymd(date)
        adj = (adj or self.adj)[0].lower()
        day_start = as_time(day_start); day_end = as_time(day_end)        
        if day_end >= day_start:
            res = self.adjust(day, adj)
            if tod > day_end: # post today trading session
                if adj == 'f':
                    if res > day:
                        return res
                    else:
                        return self.add(res, 1)
                elif adj == 'p':
                    return res
            elif tod < day_start: # pre today session
                if adj == 'f':
                    return res
                elif adj == 'p':
                    if res < day:
                        return res
                    else:
                        return self.add(res, -1)
            else: # in session. 
                return res
        else:
            if tod > day_start: # we are in tomorrow's trading session
                tomorrow = day + datetime.timedelta(1)
                res = self.adjust(tomorrow, adj)
                return res
            elif tod < day_end: # we are in today's trading session
                res = self.adjust(day , adj)
                return res
            else: # we are in post-session
                res = self.adjust(day, adj)
                if adj == 'f':
                    if res > day:
                        return res
                    else:
                        return self.add(res, 1)
                elif adj == 'p':
                    return res

    
    def adjust(self, date, adj = None):
        """

        adjust a non-business day to prev/following bussiness date

        :Parameters:
        ----------------
        date : datetime.
        adj : None or p/f/m
            adjustment convention: 'prev/following/modified following'

        :Returns:
        -------
        dateime
            nearby business day

        """
        adj = adj or self.adj
        t = ymd(date)
        t = datetime.datetime(t.year, t.month, t.day)
        if adj.startswith('f'):
            while self.is_holiday(t) and t <= self.t1:
                t = t + DAY
            while t > self.t1 and t.weekay() in self.weekend:
                t = t + DAY
            return t
                
        elif adj.startswith('p'):
            while self.is_holiday(t) and t >= self.t0:
                t = t - DAY
            while t < self.t0 and t.weekay() in self.weekend:
                t = t - DAY
            return t
        elif adj.startswith('m'):
            t = self.adjust(date, 'f')
            if t.month!=date.month:
                return self.adjust(date, 'p')
            else:
                return t

    def dt_bump(self, t, bump, adj = None):
        """
        adds a bump to a date

        :Parameters:
        ----------------
        t : datetime
            date to bump.
        bump : int, str
            bump e.g. '-1y' or '1b' or 3
        adj : adjustement type
            The default is None.

        :Returns:
        -------
        datetime
            bumped date.

        """
        if is_str(bump) and bump.lower().endswith('b'):
            return self.add(t, int(bump[:-1]))
        else:
            return dt_bump(t, bump)
        
    def clock(self, date):
        self._populate()
        date = ymd(date)
        return self.dt2int.get(date, self.dt2int[self.adjust(date)])

    def add(self, date, days, adj = None):
        """
        adjustes the start date to a business day. Then add business days on top.
        add will initiate self._populate unless we are bumping date by exactly one day

        Parameters
        ----------
        date : datetime
            start date.
        days : int
            days to bump.
        adj : str, optional
            adjusting method for date if it isn't a business day to start with.

        Returns
        -------
        datetime
            bumped date.
        """
        adj = adj or self.adj
        t = self.adjust(date, adj)
        if abs(days)>1:
            self._populate()
            return self.int2dt[self.dt2int[t] + days]
        else:
            increment = days * DAY
            res = t + increment
            while self.is_holiday(res):
                res = res + increment
        return res
    
    def bdays(self, t0, t1, adj = None):
        self._populate()
        adj = adj or self.adj
        return self.dt2int[self.adjust(t1, adj)] - self.dt2int[self.adjust(t0, adj)]
    
    def drange(self, t0 = None, t1 = None, bump = None):
        t0, t1 = self.date_range(t0, t1)
        if is_str(bump) and bump[-1] == 'b':
            self._populate()
            b = int(bump[:-1])
            i0 = self.dt2int[self.adjust(t0)]
            i1 = self.dt2int[self.adjust(t1)]
            return [self.int2dt[i] for i in range(i0, i1+b, b)]
        else:
            return drange(t0, t1, bump)
    
    def filter(self, ts, day_start = None, day_end = None):
        if day_start is None or day_end is None:
            return ts[[self.is_bday(date) for date in ts.index]]
        else:
            return ts[[self.is_trading(date, day_start = day_start, day_end = day_end) for date in ts.index]]

        
def calendar(key = None, holidays = None, weekend = None, t0 = None, t1 = None):
    """
    A function to returns either an existing calendar or construct a new one.
    - calendar('US') will return a US calendar if that is already cached
    - calendar('US', us_holiday_dates) will construct a calendar with holiday dates and then cache it
    """
    if key not in calendars or holidays is not None or weekend is not None or t0 is not None or t1 is not None:
        calendars[key] = Calendar(key, holidays = holidays, weekend = weekend, t0 = t0, t1 = t1)    
    return calendars[key]


@np.vectorize
def _weekday_intraday_clock(t):
    t = dt(t)
    j = ymd(t)
    c = calendar()._populate()
    if j in c.dt2int:
        i = t - j
        return c.dt2int[j] + (i.microseconds / 1e6 + i.seconds) / 86400
    else:
        j = c.adjust(j)        
        return c.dt2int[j]

@np.vectorize
def _fraction_clock(t):
    t = dt(t)
    return t.toordinal() + t.hour / 24 + t.minute / 1440 + (t.second + t.microsecond/1e6) / 86400

@np.vectorize
def _day_clock(t):
    t = ymd(t)
    return t.toordinal() 

@np.vectorize
def _week_clock(t):
    t = ymd(t)
    return t.toordinal() // 7

@np.vectorize
def _month_clock(t):
    t = ymd(t)
    return t.year*12 + t.month

@np.vectorize
def _year_clock(t):
    t = ymd(t)
    return t.year

@np.vectorize
def _quarter_clock(t):
    """
    >>> for m, q in zip(range(1,13), [0,0,0,1,1,1,2,2,2,3,3,3]):
    >>>     assert _quarter_clock(dt(2000,m,1)) == 8000 + q
    """
    t = dt(t)
    return t.year * 4 + (t.month-1) // 3

_clocks = dict(f = _fraction_clock, 
               d = _day_clock,
               w = _week_clock,
               m = _month_clock,
               y = _year_clock,
               q = _quarter_clock,
               k = _weekday_intraday_clock
               )


def clock(ts, time = None, t = None):
    """
    returns a vector marking the passage of time.

    :Parameters:
    ----------------
    ts : timeseries
    time : None, a string or a Calendar, or already a timeseries of times
        None: Will increment by 1 every non-nan observation
        'i' : increment by 1 every date in index (nan or not)
        'b' : weekdays distance
        'd' : day-distance (ignore intraday stamp)
        'f' : fraction-of-day-distance (do not ignore intraday stamp)
        'm' : month-distance
        'q' : quarter-distance
        'y' : year-distance
        calendar: uses the business-days distance between any two dates
        
    t: starting time in the past. 
    
    :Returns:
    -------
    an array 
        an increasing array of time such that distance between points match the above.
        
        
    :Example:
    -------
    >>> from pyg import *
    >>> assert eq(clock(pd.Series(np.arange(10), drange(2000, 9))), np.arange(1,11))
    >>> assert eq(clock(pd.Series(np.arange(10), drange(2000, 9)), t = 5), np.arange(6,16))
    >>> assert eq(clock(pd.Series(np.arange(10), drange(2000, 9)), 'i'), np.arange(1,11))
    >>> assert eq(clock(pd.Series(np.arange(10), drange(2000, 9)), 'b'), np.array([26090, 26090, 26090, 26091, 26092, 26093, 26094, 26095, 26095, 26095]))
    >>> assert eq(clock(pd.Series(np.arange(10), drange(2000, '9b', '1b')), 'b'), np.arange(26090, 26100))

    >>> assert eq(clock(np.arange(10)), np.arange(1,11))
    >>> assert eq(clock(pd.Series(np.arange(10)), t = 5), np.arange(6,16))
    >>> assert eq(clock(np.arange(10), 'i'), np.arange(1,11))

    """
    if t is None or is_nan(t):
        t = 0
    if is_str(time):
        time = time[0].lower()
    if is_ts(time) or is_arr(time):
        return time
    elif time is None and (is_arr(ts) or is_pd(ts)):
        return np.full(len(ts), np.nan)
        # mask = np.min(np.isnan(ts), axis = 1) if len(ts.shape) == 2 else np.isnan(ts)
        # res = np.full(len(ts), 1)
        # res[mask] = 0
        # return t + np.cumsum(res)        
    elif time == 'i' and (is_arr(ts) or is_pd(ts)): #index
        return t + np.arange(1,1+len(ts))
    elif is_ts(ts):    
        if isinstance(time, Calendar):
            return np.vectorize(time.clock)(ts.index)
        elif callable(time):
            return np.vectorize(time)(ts.index)            
        elif time == 'b':
            time = calendar()
            return np.vectorize(time.clock)(ts.index)
        elif time in _clocks:
            return _clocks[time](ts.index)
        else:
            raise ValueError('cannot determine time for ts=%s, time=%s'%(ts, time))            
    else:
        raise ValueError('cannot determine time for ts=%s, time=%s'%(ts, time))




import pandas as pd; import numpy as np
from pyg_base._inspect import getargs
from pyg_base._types import is_ts, is_str, is_df, is_pd, is_series, is_array, is_tuple, is_dict
from pyg_base._decorators import wrapper
from pyg_base._as_list import as_tuple

__all__ = ['loop', 'loops', 'len0', 'pd2np', 'shape', 'loop_all']


def _zero():
    return 0

def len0(value):
    """
    returns the len of an object or 0 if no len(value) exists
    
    :Example:
    ---------
    >>> assert len0(5) == 0
    >>> assert len0('a string is dimensionless') == 0
    >>> assert len0([]) == 0
    >>> assert len0([1,2]) == 2

    Parameters
    ----------
    value : TYPE
        DESCRIPTION.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """
    try:
        return 0 if is_str(value) else getattr(value, '__len__', _zero)()
    except Exception:
        return 0


def shape(value):
    return getattr(value, 'shape', ())
 

def _item_by_key(value, key, keys, i = None):
    if isinstance(value, dict) and sorted(value.keys()) == keys:
        return value[key] 
    elif isinstance(value, pd.Series) and sorted(value.index.values) == keys:
        return value[key] 
    elif isinstance(value, pd.DataFrame) and sorted(value.columns) == keys:
        return value[key]
    elif is_array(value) and len(value.shape):
        if len(value.shape) == 2 and value.shape[1] == len(keys) and i is not None:
            return value.T[i]
        elif len(value.shape) == 1 and value.shape[0] == len(keys) and i is not None:
            return value[i]
        else:
            return value
    else:
        if i is not None and len0(value) == len(keys):
            return value[i]
        else:
            return value

def _item_by_i(value, i, n):
    if is_df(value) and value.shape[1] == 1:
        value = value.iloc[:,0]
    elif is_array(value) and len(value.shape) == 2 and value.shape[1] == 1:
        value = value.T[0]
    if isinstance(value, (list, tuple)) and len(value) == n:
        return value[i]
    elif is_array(value):
        if len(value.shape) == 2 and value.shape[-1] == n:
            return value.T[i]
        if len(value.shape) == 1 and value.shape[-1] == n:
            return value[i]
    elif is_series(value) and len(value) == n and not is_ts(value):
        return value.iloc[i]
    elif is_df(value) and value.shape[1] == n:
        return value.iloc[:, i]
    return value           

# add_index_and_columns = _np2pd

def add_index_and_columns(res, arg):
    if isinstance(res, tuple):
        return tuple([add_index_and_columns(r, arg) for r in res])
    elif isinstance(res, dict):
        return type(res)({k: add_index_and_columns(r, arg) for k,r in res.items()})
    if isinstance(res, (pd.Series, pd.DataFrame)) and len(res) == len(arg):
        res.index = arg.index
    if isinstance(res, (pd.DataFrame)) and res.shape[1] == arg.shape[1]:
        res.columns = arg.columns
    return res

def axis0_to_dataframe(res, arg):
    if min([is_tuple(r) for r in res], default = False) and len(set([len(r) for r in res])) == 1:
        return tuple([axis0_to_dataframe(r, arg) for r in list(zip(*res))])
    elif min([is_dict(r) for r in res], default = False) and len(set([len(r) for r in res])) == 1:
        keys = res[0].keys()
        tp = type(res[0])
        return tp({key : axis0_to_dataframe([r[key] for r in res], arg) for key in keys})
    try:
        res = type(arg)(dict(zip(arg.columns, res)))
        return add_index_and_columns(res, arg)
    except ValueError:
        res = pd.Series(res)
        res.index = arg.columns
        return res

def axis0_to_array(res, arg):
    if min([is_tuple(r) for r in res], default = False) and len(set([len(r) for r in res])) == 1:
        return tuple([axis0_to_array(r, arg) for r in list(zip(*res))])
    elif min([is_dict(r) for r in res], default = False) and len(set([len(r) for r in res])) == 1:
        keys = res[0].keys()
        tp = type(res[0])
        return tp({key : axis0_to_array([r[key] for r in res], arg) for key in keys})
    rtn = np.array(res)
    if len(rtn.shape)>1:
        rtn = rtn.T
    return rtn

class loops(wrapper):

    """
    
    converts a function to loop over the arguments, depending on the type of the first argument

    :Examples:
    --------------
    >>> @loop(dict, list, pd.DataFrame, pd.Series)
    >>> def f(a,b):
    >>>     return a+b
    
    >>> assert f(1,2) == 3
    >>> assert f([1,2,3],2) == [3,4,5]
    >>> assert f([1,2,3], [4,5,6]) == [5,7,9]    
    
    >>> assert f(dict(x=1,y=2), 3) == dict(x = 4, y = 5)
    >>> assert f(dict(x=1,y=2), dict(x = 3, y = 4)) == dict(x = 4, y = 6)
    

    >>> a = pd.Series(dict(x=1,y=2))
    >>> b = dict(x=3,y=4)
    >>> assert np.all(f(a,b) == pd.Series(dict(x=4,y=6)))    

    >>> a = pd.DataFrame(dict(x=[1,1],y=[2,2])); a.index = [5,10]
    >>> b = dict(x=3,y=4)
    >>> res =  f(a,b)
    >>> assert np.all(res == pd.DataFrame(dict(x=[4,4],y=[6,6]), index = [5,10]))    
    
    >>> a = pd.DataFrame(dict(x=[1,1],y=[2,2])); a.index = [5,10]
    >>> res =  f(a,[3,4])
    >>> assert np.all( res == pd.DataFrame(dict(x=[4,4],y=[6,6]), index = [5,10]))    
    
    
    """

    def __init__(self, function = None, types = None):
        super(loops, self).__init__(function = function, types = as_tuple(types))

    @property
    def first(self):
        args = getargs(self)
        return args[0] if len(args) else None
    
    def T(self, arg, args, kwargs):
        arg_ = _T(arg)
        args_ = _T(args)
        kwargs_ = _T(kwargs)
        res = self._wrapped(arg_, args_, kwargs_)
        return _T(res)

    def wrapped(self, *args, **kwargs):
        top = self.first
        if len(args) == 0 and not top in kwargs:
            return self.function(*args, **kwargs)
        if len(args):
            arg, args_, kwargs_ = args[0], args[1:], kwargs
        else:
            arg = kwargs.pop(top)
            args_, kwargs_ = args, kwargs
        if isinstance(arg, pd.Series) and pd.Series in self.types and not is_ts(arg):
            keys = sorted(arg.index)
            res = {key : self._wrapped(arg[key], (_item_by_key(a,key,keys) for a in args_), {k : _item_by_key(v,key,keys) for k,v in kwargs_.items()}) for key in arg.index}
            return type(arg)(res)          
        else:
            return self._wrapped(arg, args_, kwargs_)
    
    def _wrapped(self, arg, args, kwargs):
        axis = kwargs.pop('axis', 0)
        if isinstance(arg, dict) and dict in self.types:
            keys = sorted(arg.keys())
            res = {key : self._wrapped(arg[key], (_item_by_key(a,key,keys) for a in args), {k : _item_by_key(v,key,keys) for k,v in kwargs.items()}) for key in arg.keys()}
            return type(arg)(res)
        elif isinstance(arg, pd.DataFrame) and pd.DataFrame in self.types:
            if axis in (1,-1):
                res = self.T(arg, args, kwargs)
                if isinstance(res, (pd.DataFrame, pd.Series)) and len(res) == len(arg):
                    res.index = arg.index
                return add_index_and_columns(res, arg)
            else:
                keys = sorted(arg.columns)
                res = [self._wrapped(arg[key], (_item_by_key(a,key,keys,i) for a in args), {k : _item_by_key(v,key,keys,i) for k,v in kwargs.items()}) for i, key in enumerate(arg.columns)]
                rtn = axis0_to_dataframe(res, arg)
                return rtn
        elif isinstance(arg, pd.Series):
            return self.function(arg, *args, **kwargs)
        elif isinstance(arg, np.ndarray) and np.ndarray in self.types:
            if len(arg.shape) <= 1:
                return self.function(arg, *args, **kwargs) 
            else:
                if axis in (1,-1):
                    return self.T(arg, args, kwargs)
                else:
                    n = arg.shape[-1]
                    res = [self._wrapped(_item_by_i(arg,i,n), (_item_by_i(a,i,n) for a in args), {k: _item_by_i(v,i,n) for k, v in kwargs.items()}) for i in range(n)]
                    return axis0_to_array(res, arg)
        elif isinstance(arg, self.types):
            n = len(arg)
            res = [self._wrapped(arg[i], (_item_by_i(a,i,n) for a in args), {k: _item_by_i(v,i,n) for k, v in kwargs.items()}) for i in range(n)]                            
            return type(arg)(res)
        else:
            return self.function(arg, *args, **kwargs)

def loop(*types):
    """
    returns an instance of loops(types = types)
    """
    return loops(types = as_tuple(types))


def _T(arg):
    if isinstance(arg, tuple):
        return type(arg)([_T(a) for a in arg])
    elif isinstance(arg, list):
        return type(arg)([_T(a) for a in arg])
    elif isinstance(arg, dict):
        return type(arg)({k :_T(a) for k,a in arg.items()})
    else:
        return arg.T if isinstance(arg, (np.ndarray, pd.DataFrame)) and len(arg.shape) == 2 else arg

def _values(a):
    if is_series(a):
        return a.values
    elif is_df(a):
        if a.shape[-1] == 1:
            return a.values.T[0]
        else:
            return a.values
    else:
        return a

def _np2pd(res, arg):
    if isinstance(res, tuple):
        return tuple([_np2pd(r, arg) for r in res])
    elif isinstance(res, tuple):
        return [_np2pd(r, arg) for r in res]
    elif isinstance(res, dict):
        return type(res)({k: _np2pd(r, arg) for k,r in res.items()})
    n = len0(res)
    if n == len0(arg):
        if len(res.shape) == 2:
            res = pd.DataFrame(res, arg.index)
            if len(arg.shape) == 2 and res.shape[1] == arg.shape[1]:
                res.columns = arg.columns
            return res
        else:
            return pd.Series(res, arg.index)
    elif is_df(arg) and n == arg.shape[1]:
        return pd.Series(arg.columns, res)
    else:
        return res

class pd2np(wrapper):
    """
    converts a numpy-based function to work on pandas by converting to a numpy arrage and then converting back to panadas dataframe
    """
    def wrapped(self, *args, **kwargs):
        if len(args):
            arg = args[0]
        else:
            arg = kwargs[getargs(self.function)[0]]
        if not is_pd(arg):
            return self.function(*args, **kwargs)
        args_ = [_values(a) for a in args]
        kwargs_ = {k : _values(v) for k, v in kwargs.items()}
        res = self.function(*args_, **kwargs_)
        return _np2pd(res, arg)


loop_all = loops(types = (pd.Series, pd.DataFrame, np.ndarray, list, tuple, dict))




from pyg_base._as_float import as_float
from pyg_base._as_list import as_list, as_tuple, first, last, passthru, unique, is_rng
from pyg_base._acell import acell, acell_load, acell_func, acell_go
from pyg_base._cache import cache, get_cache
from pyg_base._cell import cell, cell_go, cell_item, cell_func, cell_load, cell_output, cell_clear, cell_inputs
from pyg_base._cfg import mkdir, cfg_read, cfg_write, CFG
from pyg_base._dates import dt,dt_bump, as_tz, tz_info, today, ymd, TMIN, TMAX, DAY, futcodes, dt2str, is_bump, nth_weekday_of_month
from pyg_base._dag import get_DAG, get_DAGs, get_GADs, get_GAD, add_edge, topological_sort, descendants, del_edge
from pyg_base._decorators import kwargs_support, wrapper, try_value, try_back, try_nan, try_none, try_zero, try_false, try_true, try_list, timer
from pyg_base._dict import Dict, items_to_tree, tree_items, tree_keys, tree_values, tree_setitem, tree_getitem, tree_get, tree_update, dict_invert
from pyg_base._dictattr import dictattr, relabel
from pyg_base._dictable import dictable, dict_concat, is_dictable
from pyg_base._drange import date_range, drange, calendar, Calendar, clock, as_time
from pyg_base._encode import encode, decode, bson2pd, pd2bson, as_primitive
from pyg_base._eq import eq, in_
from pyg_base._file import read_csv
from pyg_base._getitem import getitem, callitem, callattr
from pyg_base._inspect import getargspec, getargs, getcallargs, call_with_callargs, argspec_defaults, argspec_required, argspec_update, argspec_add, kwargs2args
from pyg_base._logger import logger, get_logger
from pyg_base._loop import loop, loops, len0, pd2np, shape, loop_all
from pyg_base._named_dict import named_dict
from pyg_base._pandas import df_index, df_columns, df_reindex, np_reindex, df_concat, df_fillna, presync, df_sync, \
    add_, sub_, mul_, div_, pow_, df_slice, df_unslice, nona, min_, max_, gt_, ge_, lt_, le_
from pyg_base._parquet import pd_to_parquet, pd_read_parquet
from pyg_base._perdictable import join, perdictable
from pyg_base._reducer import reducer, reducing
from pyg_base._sort import sort, cmp, Cmp
from pyg_base._tree import is_tree, tree_to_table
from pyg_base._tree_repr import tree_repr
from pyg_base._txt import alphabet, ALPHABET, f12, replace, lower, upper, proper, strip, split, capitalize, common_prefix, deprefix
from pyg_base._types import is_pd, is_arr, is_int, is_float, is_num, is_bool, is_str, is_nan, \
    is_none, is_dict, is_iterable, is_date, is_df, is_series, is_ts, is_list, is_tuple,\
    is_pds, is_arrs, is_ints, is_floats, is_nums, is_len, is_bools, is_strs, is_nans, is_nones, is_dicts, is_zero_len, \
    is_iterables, is_tss, nan2none, NoneType, is_dates, is_lists
from pyg_base._ulist import ulist, rng
from pyg_base._waiter import waiter, async_wrapper
from pyg_base._zip import zipper, lens
from pyg_npy import np_save, pd_to_npy, pd_read_npy, path_name, path_dirname, path_join, mkdir

